""" Solver Formatting Functions.
This file provides easier to use functions to interface with in creating variables and constraints for the optimiser.
"""
import pandas as pd
import numpy as np

def create_timeseries_vars(self, var_name, lb, ub, solver_type='continuous'):
	"""
	Creates a series of variables, one for each interval timestamp. Variable is defined by a name, lower bound, upper
	bound and variable type.

	Parameters
	----------
	var_name : str
		Variable name, used to reference and call variable within the optimiser.
	lb : float
		Lower bound parameter for minimum possible value of variable.
	ub : float
		Upper bound parameter for maximum possible value of variable.
	solver_type : ['binary','continuous'], optional
		Variable type. The default is 'continuous'.
	"""
	new_vars = pd.DataFrame({'variable_id': range(self._varid_idx,self._varid_idx+self._n),'interval': range(self._n),
						  'lower_bound': lb, 'upper_bound': ub, 'type': solver_type})
	self._varid_idx = max(new_vars['variable_id']) + 1
	self._var[var_name] = new_vars


def create_period_vars(self, var_name, interval_set, lb, ub, solver_type='continuous'):

	new_vars = pd.DataFrame({'variable_id': range(self._varid_idx, self._varid_idx+len(interval_set)), \
		'interval': interval_set['interval'], 'lower_bound': lb, 'upper_bound': ub, 'type': solver_type})
	self._varid_idx = max(new_vars['variable_id']) + 1
	self._var[var_name] = new_vars


def create_var(self, var_name, lb, ub, solver_type='continuous'):
	"""
	Create a single variable. Variable is defined by a name, lower bound, upper bound and variable type.

	Parameters
	----------
	var_name : str
		Variable name, used to reference and call variable within the optimiser.
	lb : float
		Lower bound parameter for minimum possible value of variable.
	ub : float
		Upper bound parameter for maximum possible value of variable.
	solver_type : ['binary','continuous'], optional
		Variable type. The default is 'continuous'.
	"""
	new_var = pd.DataFrame({'variable_id': range(self._varid_idx,self._varid_idx+1),'interval': None,
						  'lower_bound': lb, 'upper_bound': ub, 'type': solver_type})
	self._varid_idx = max(new_var['variable_id']) + 1
	self._var[var_name] = new_var    

def create_constr_rhs(self, constr_name, constr_type='<=', rhs_value=0.0):
	"""
	Create the right-hand side of a single constrant. This generates a unique constraint ID for the new constraint. 
	Constrant RHS is defined by a name, contraint/inequality type and RHS value.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_type : ['<=','==','>=']
		In/equality type for the constraint. The default is '<='.
	rhs_value : float
		Defined static value for the RHS of the constraint. The default is 0.0.
	"""
	new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+1),'interval': None,
							'type': constr_type, 'rhs': rhs_value})
	self._conid_idx = max(new_constr['constraint_id']) + 1
	self._constr_rhs[constr_name] = new_constr    

def create_constr_rhs_on_interval(self, constr_name, constr_type='<=', rhs_value=0.0, intervals=[]):
	"""
	Create the right-hand side for a set of new constraints corresponding to each interval. This generates unique 
	constraint IDs for consequtive intervals. Constrant RHS is defined by a name, contraint/inequality type and RHS
	value.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_type : ['<=','==','>=']
		In/equality type for the constraint. The default is '<='.
	rhs_value : float
		Defined static value for the RHS of the constraint. The default is 0.0.
	"""
	if intervals == []:
		int_list = range(self._n)
	else:
		int_list = intervals

	if (type(rhs_value) == int) or (type(rhs_value) == float):
		new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+len(int_list)),
							 'interval': int_list, 'type': constr_type, 'rhs': rhs_value})
	elif type(rhs_value) == list:
		new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+len(int_list)),
							 'interval': int_list, 'type': constr_type, 'rhs': rhs_value})
	else:
		raise ValueError("Invalid rhs_value type")
	self._conid_idx = max(new_constr['constraint_id']) + 1
	self._constr_rhs[constr_name] = new_constr

def create_constr_rhs_on_interval_dynamicvar(self, constr_name, constr_type, rhs_var_id_series, intervals=[]):
	"""
	Create the right-hand side for a set of new constraints corresponding to each interval with a dynamic RHS variable.
	The RHS is defined by a variable ID corresponding to each interval. This function generates unique 
	constraint IDs for consequtive intervals. Constrant RHS is defined by a name, contraint/inequality type and series
	of variable IDs for the RHS.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_type : ['<=','==','>=']
		In/equality type for the constraint. The default is '<='.
	rhs_var_id_series : pd.DataFrame()
		Dataframe containing interval column and variable_id column.
	"""
	if intervals == []:
		int_list = range(self._n)
	else:
		int_list = intervals

	new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+len(int_list)),
							   'interval': int_list, 'type': constr_type})
	new_constr = new_constr.merge(rhs_var_id_series, on=['interval'])
	new_constr.rename(columns={'variable_id':'rhs_variable_id'},inplace=True)
	self._conid_idx = max(new_constr['constraint_id']) + 1
	self._constr_rhs_dynamic[constr_name] = new_constr

def create_constr_rhs_dynamicvar(self, constr_name, constr_type, rhs_var_id):
	"""
	Create the right-hand side as a dynamic variable ID for a single new constraint. This function generates a unique 
	constraint ID. Constrant RHS is defined by a name, contraint/inequality type and variable ID integer for the RHS.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_type : ['<=','==','>=']
		In/equality type for the constraint. The default is '<='.
	rhs_var_id : int
		Variable ID for parameter to sit on the RHS.
	"""
	new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+1),'interval': None,
							'type': constr_type, 'rhs_variable_id': rhs_var_id})
	self._conid_idx = max(new_constr['constraint_id']) + 1
	self._constr_rhs_dynamic[constr_name] = new_constr

def create_constr_lhs_sum_intervals(self, constr_name, constr_rhs, coefficient_map=1):
	"""
	** IN-DEVELOPMENT **
	Create definitions for the LHS of a defined single constraint (by a previous RHS formulation), mapping a set of 
	variables for all intervals to be summated on the LHS with coefficients of 1. The RHS arugment must already be
	defined.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_rhs : pd.DataFrame()
		The dataframe entry for the corresponding RHS constraint. In the form of, self._constr_rhs[{var_name}]
	coefficient_map : int
		The LHS coefficient mapped to each LHS variable. The default is 1.
	"""
	new_lhs = pd.DataFrame()
	for var in coefficient_map:
		coeff_df = pd.DataFrame({'constraint_id': constr_rhs['constraint_id'].values[0],
							   'variable_id': self._var[var]['variable_id'],
							   'coefficient': coefficient_map[var]})
		new_lhs = pd.concat([new_lhs, coeff_df])
	self._constr_lhs[constr_name] = new_lhs


def create_constr_lhs_sum_up_to_intervals(self, constr_name, constr_rhs, coefficient_map):
	
	new_lhs = pd.DataFrame()
	prev_itval_id = -1
	for idx, itval_id in enumerate(constr_rhs['interval']):
		for var in coefficient_map:
			sel_var_ids = self._var[var][(self._var[var]['interval'] > prev_itval_id) & \
				(self._var[var]['interval'] <= itval_id)]
			coeff_df = pd.DataFrame({'constraint_id': constr_rhs['constraint_id'][idx],
								'variable_id': sel_var_ids['variable_id'],
								'coefficient': coefficient_map[var]})
			new_lhs = pd.concat([new_lhs, coeff_df])
		prev_itval_id = itval_id
	self._constr_lhs[constr_name] = new_lhs


def create_constr_lhs_on_interval(self, constr_name, constr_rhs, coefficient_map):
	"""
	Create definitions for the LHS of a defined set of RHS constraints corresponding to each interval (by previous RHS 
	formulation). The mapping of LHS variables and coefficients may either be a single variable applied to all intervals
	of the RHS constraint set, or a different (interval-specific) variable corresponding to interval-specific RHS
	constraints. The coefficient map parameter allows multiple LHS variables to be considered in a single function call.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_rhs : pd.DataFrame()
		The dataframe entry for the corresponding RHS constraint. In the form of, self._constr_rhs[{constr_name}]
	coefficient_map : dict
		Defines the matching coefficients to each variable ID set. The same coefficient must be applied to all LHS
		variables of the same name for all intervals. In the form of, {'{var_name_1}': 1, '{var_name_2}': -1}
	"""
	new_lhs = pd.DataFrame()
	for var in coefficient_map:
		# Collect rhs terms
		rhs_df = constr_rhs[['constraint_id','interval']] 
		# Collect lhs terms with coefficient
		lhs_coef = self._var[var][['variable_id','interval']]
		lhs_coef.insert(2,'coefficient',coefficient_map[var])
		# Merge lhs terms to rhs terms
		if (len(lhs_coef) == 1) and (lhs_coef['interval'].isnull().any()): # apply single var to all intervals
			rhs_df.insert(2,'variable_id',lhs_coef.loc[0,'variable_id'])
			rhs_df.insert(3,'coefficient',lhs_coef.loc[0,'coefficient'])
			new_lhs = pd.concat([new_lhs, rhs_df])
		else:
			new_lhs = pd.concat([new_lhs, rhs_df.merge(lhs_coef, on=['interval'])])
	
	if constr_name in self._constr_lhs:
		self._constr_lhs[constr_name] = pd.concat([self._constr_lhs[constr_name], new_lhs])
	else:
		self._constr_lhs[constr_name] = new_lhs

def create_constr_lhs_on_interval_dynamic(self, constr_name, constr_rhs, decision_vars, coefficient_list):
	"""
	Create definitions for LHS variable with a dynamic coefficient across intervals. For example used to define VRE
	availability where LHS arguments are dynamic coefficient (renewable trace) * single-variable across intervals (VRE 
	capacity) == RHS dynamic variable (VRE availability)

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	constr_rhs : pd.DataFrame
		The dataframe entry for the corresponding RHS constraint. In the form of, self._constr_rhs[{var_name}]
	decision_vars : pd.DataFrame
		The dataframe of variables to be used as LHS arguments. In the form of, self._var[name_cap].
	coefficient_list : pd.Series
		Dynamic values to be applied as coefficients, ordered by consequtive intervals.

	Raises
	------
	NotImplementedError
		If the decision variable dataframe passed has variables for distinct intervals this error will raise. Only a 
		single variable with non-descript interval data can be passed as decision_vars.
	"""
	new_lhs = constr_rhs[['constraint_id','interval']]

	if not (len(decision_vars['interval']==1) & (decision_vars['interval'].values[0] == None)):
		new_lhs = new_lhs.merge(decision_vars[['variable_id','interval']], on='interval')
	else:
		lhs_var = decision_vars['variable_id'].values[0]
		new_lhs.insert(2,'variable_id',lhs_var)

	new_lhs.insert(3,'coefficient',coefficient_list)
	self._constr_lhs[constr_name] = new_lhs

def create_constr_lhs_on_interval_fixed_rhs(self, constr_name, constr_rhs, decision_vars, coefficient_map):
	"""
	** IN DEVELOPMENT ** 
	```{note}
	Redundancy with <create_constr_lhs_on_interval()>. Remove expression in future bug fixes.
	```
	"""
	new_lhs = pd.DataFrame()
	for var in coefficient_map:
		lhs_coef = decision_vars[var][['variable_id','interval']]
		lhs_coef.insert(2,'coefficient',coefficient_map[var])
		lhs_coef.insert(0,'constraint_id',constr_rhs['constraint_id'].values[0])
		new_lhs = pd.concat([new_lhs, lhs_coef])
	self._constr_lhs[constr_name] = new_lhs
	#return new_lhs


def create_constr_lhs_ramp(self, constr_name, constr_rhs, coeff_map_curr, coeff_map_next):

	new_lhs = pd.DataFrame()
	# Mapping coefficients for current intervals
	for var in coeff_map_curr:
		rhs = constr_rhs.copy()
		rhs = rhs[['constraint_id','interval']]
		lhs_coef = self._var[var][['variable_id','interval']]
		lhs_coef.insert(2,'coefficient', coeff_map_curr[var])
		lhs_coef = rhs.merge(lhs_coef, on='interval')
		new_lhs = pd.concat([new_lhs, lhs_coef])

	# Mapping coefficients for next intervals
	for var in coeff_map_next:
		rhs = constr_rhs.copy()
		rhs = rhs[['constraint_id','interval']]
		rhs.insert(2,'next',rhs['interval'] + 1) # NEW

		lhs_coef = self._var[var][['variable_id','interval']]
		lhs_coef.insert(2,'coefficient', coeff_map_next[var])
		lhs_coef.rename(columns={'interval': 'next'},inplace=True) # NEW
		lhs_coef = rhs.merge(lhs_coef, on='next') # DIFFERENT
		new_lhs = pd.concat([new_lhs, lhs_coef[lhs_coef.columns[lhs_coef.columns != 'next']]])

	new_lhs.to_csv(rf'E:\PROJECTS\NEMGLO\localonly\lhs_{constr_name}.csv')
	self._constr_lhs[constr_name] = new_lhs

	return


def create_contr_lhs_on_interval_matrix(self, constr_name, constr_rhs, coeff_matrix):
	"""
	
	|		| h2e_load	| h2_production	|
	|	-2	|	NaN		|	1.0			|
	|	-1	|	-1.0	|	NaN			|

	The above matrix index indicates the interval offset to apply for that row of variables, the values within the coefficents.
	"""
	new_lhs = pd.DataFrame()
	coeff_matrix = coeff_matrix.replace({np.nan: None})

	# Mapping coefficients based on matrix index
	for row, col in [(x, y) for x in coeff_matrix.index for y in coeff_matrix.columns]:
		if coeff_matrix.loc[row, col] is not None:
			rhs = constr_rhs.copy()
			rhs = rhs[['constraint_id','interval']]
			rhs.insert(2,'match_offset',rhs['interval'] + row) # Row is the offset to shift interval series

			lhs_coef = self._var[col][['variable_id','interval']]
			lhs_coef.insert(2,'coefficient', coeff_matrix.loc[row, col])
			lhs_coef = lhs_coef.rename(columns={'interval': 'match_offset'})
			lhs_coef = rhs.merge(lhs_coef, on='match_offset')
			new_lhs = pd.concat([new_lhs, lhs_coef[lhs_coef.columns[lhs_coef.columns != 'match_offset']]])

	# new_lhs.to_csv(rf'E:\PROJECTS\NEMGLO\localonly\lhs_{constr_name}.csv')
	self._constr_lhs[constr_name] = new_lhs




def create_constr_bigM_on_interval(self, constr_name, lhs_vars, rhs_bin_vars, constr_type='<=', mode='inverse', 
								   Mvalue=10000.0):
	"""
	Creates both the LHS and RHS elements in a bigM constraint formulation in one function.

	Parameters
	----------
	constr_name : str
		Constraint name, used to reference and call constraint within the optimiser.
	lhs_vars : pd.DataFrame
		The dataframe of LHS variables (interval-specific) to form the constraint. In the form of,
		self._var['{var_name}'].
	rhs_bin_vars : pd.DataFrame
		The dataframe of RHS variables (interval-specific) to form the constraint. In the form of,
		self._var['{var_name}'].
	constr_type : ['<=','==','>=']
		In/equality type for the constraint. The default is '<='.
	mode : ['inverse','normal']
		Determines the structure of the bigM constraint formulation. 'inverse' structures the formulation as:
		LHS <= M(1-RHS). 'normal' structures the formulation as: LHS <= M(RHS). The default is 'inverse'.
	Mvalue : float, optional
		The magnitude of the bigM 'swamping' parameter. The default is 10000.0.

	Raises
	------
	ValueError
		Raised if mode parameter is passed as something other than 'inverse' or 'normal'.
	"""
	if not ((mode == 'inverse') or (mode == 'normal')):
		raise ValueError('Unrecognised mode variable. Must be: normal, inverse')

	new_constr = pd.DataFrame({'constraint_id': range(self._conid_idx,self._conid_idx+self._n),
							'interval': range(self._n), 
							'constr_type': constr_type,
							'mode': mode,
							'bigM': Mvalue})

	lhs = lhs_vars[['variable_id','interval']]
	new_constr = new_constr.merge(lhs, on=['interval'])
	new_constr.rename(columns={'variable_id':'lhs_variable_id'},inplace=True)
	rhs = rhs_bin_vars[['variable_id','interval']]
	new_constr = new_constr.merge(rhs, on=['interval'])
	new_constr.rename(columns={'variable_id':'rhs_variable_id'},inplace=True)
	
	self._conid_idx = max(new_constr['constraint_id']) + 1
	self._constr_bigM[constr_name] = new_constr

def create_objective_cost(self, var_name, decision_var_series, cost=1000.0):
	"""
	Creates an objective cost parameter corresponding to a decision variable or set of decision variables. If the cost
	is passed as a float, the single value is mapped to all decision variables. However, if passed as a pd.DataFrame
	with a defined interval column, the unique interval-specific costs are mapped corresponding to interval-specific
	variables.

	Parameters
	----------
	var_name : str
		Variable name, used to reference and call variable within the optimiser.
	decision_var_series : pd.DataFrame
		The dataframe containing the decision variables to be costed in the objective function. In the form of, 
		self._var['{var_name}'].
	cost : float
		The corresponding cost mapped to the decision variables passed. The default is 1000.0.
	"""
	new_costs = decision_var_series[['variable_id','interval']]
	if (type(cost) == int) or (type(cost) == float):
		new_costs.insert(2,'cost', cost)
	elif type(cost) == pd.DataFrame:
		new_costs = new_costs.merge(cost, on=['interval'])
	self._objective_cost[var_name] = new_costs

def create_sos_type_2(self, x_samples, y_samples, sos_name, weight_var_name, x_var_name, y_var_name):

	# Arrange weight variable df
	weights_df = pd.concat([self._var[k] for k in self._var.keys() if k.__contains__(weight_var_name)], \
		ignore_index=True)
	weights_df = weights_df[['variable_id','interval']]

	x_var_df = self._var[x_var_name][['variable_id','interval']]
	y_var_df = self._var[y_var_name][['variable_id','interval']]

	self._sos_2[sos_name] = {
		'weights': weights_df,
		'x_samples': x_samples,
		'y_samples': y_samples,
		'x_vars': x_var_df,
		'y_vars': y_var_df
	}


def remove_element(plan, name, all_assoc=True, as_var=False, as_objective_cost=False, as_constr_lhs=False, \
	as_constr_rhs=False, as_constr_rhs_dynamic=False, as_constr_bigM=False, as_sos_2=False):
	"""Remove variable and/or constraint elements from the Plan object parsed.

	Parameters
	----------
	plan : nemglo.planner.Plan
		Plan object to remove variable/constraints from.
	name : str
		Variable or constraint name to remove. The identifier of a component object may also be used to remove that
		component entirely using all_assoc.
	all_assoc : bool, optional
		Set as true to remove all associations of the name from the Plan object, in all variable and constraint
		attributes.  
	as_var : bool, optional
		Set as true to remove variable name from the variable attributes of Plan object, by default False.
	as_objective_cost : bool, optional
		Set as true to remove variable name from the objective cost attributes of Plan object, by default False.
	as_constr_lhs : bool, optional
		Set as true to remove constraint name from the LHS constraint attributes of Plan object, by default False.
	as_constr_rhs : bool, optional
		Set as true to remove constraint name from the RHS constraint attributes of Plan object, by default False
	as_constr_rhs_dynamic : bool, optional
		Set as true to remove constraint name from the Dynamic RHS constraint attributes of Plan object, by default
		False.
	as_constr_bigM : bool, optional
		Set as true to remove constraint name from the Big M constraint attributes of Plan object, by default False.
	as_sos_2 : bool, optional
		Set as true to remove constraint name from the SOS Type 2 constraint attributes of Plan object, by default
		False.
	"""
	if as_var or all_assoc:
		for item in [k for k in plan._var.keys() if k.__contains__(name)]:
			plan._var.pop(item)
			print("Plan '{}' removed variable '{}'".format(plan._id, item))

	if as_objective_cost or all_assoc:
		for item in [k for k in plan._objective_cost.keys() if k.__contains__(name)]:
			plan._objective_cost.pop(item)
			print("Plan '{}' removed objective cost '{}'".format(plan._id, item))

	if as_constr_lhs or all_assoc:
		for item in [k for k in plan._constr_lhs.keys() if k.__contains__(name)]:
			plan._constr_lhs.pop(item)
			print("Plan '{}' removed constraint LHS '{}'".format(plan._id, item))

	if as_constr_rhs or all_assoc:
		for item in [k for k in plan._constr_rhs.keys() if k.__contains__(name)]:
			plan._constr_rhs.pop(item)
			print("Plan '{}' removed constraint RHS '{}'".format(plan._id, item))

	if as_constr_rhs_dynamic or all_assoc:
		for item in [k for k in plan._constr_rhs_dynamic.keys() if k.__contains__(name)]:
			plan._constr_rhs_dynamic.pop(item)
			print("Plan '{}' removed constraint RHS dynamic '{}'".format(plan._id, item))

	if as_constr_bigM or all_assoc:
		for item in [k for k in plan._constr_bigM.keys() if k.__contains__(name)]:
			plan._constr_bigM.pop(item)
			print("Plan '{}' removed objective cost '{}'".format(plan._id, item))

	if as_sos_2 or all_assoc:
		for item in [k for k in plan._sos_2.keys() if k.__contains__(name)]:
			plan._sos_2.pop(item)
			print("Plan '{}' removed objective cost '{}'".format(plan._id, item))

	if all_assoc:
		if len([k for k in plan._components if k._id.__contains__(name)]) > 0:
			plan._components.remove([k for k in plan._components if k._id.__contains__(name)][0])
			print("Plan '{}' unlinked the component '{}'".format(plan._id, name))
		else:
			print("Plan '{}' does not have any component '{}' to remove".format(plan._id, name))









import plotly.io as pio

PALETTEZZZ = ['#b2d4ee','#849db1','#4f6980',
           '#B4E3BC','#89AE8F','#638b66',
           '#ffb04f','#de9945','#af7635',
           '#ff7371','#d6635f','#b65551',
           '#AD134C','#cc688d','#ff82b0']

NORD = {
    'polar_1': '#4c566a',
    'polar_2': '#434c5e',
    'polar_3': '#3b4252',
    'polar_4': '#2e3440',
    'snow_1': '#eceff4',
    'snow_2': '#e5e9f0',
    'snow_3': '#d8dee9',
    'frost_1': '#5e81ac',
    'frost_2': '#81a1c1',
    'frost_3': '#88c0d0',
    'frost_4': '#8fbcbb',
    'aurora_p': '#b48ead',
    'aurora_g': '#a3be8c',
    'aurora_y': '#ebcb8b',
    'aurora_o': '#d08770',
    'aurora_r': '#bf616a',
}

PAL22 = {
    'blue_1': '#b2d4ee',
    'blue_2': '#849db1',
    'blue_3': '#4f6980',
    'green_1': '#B4E3BC',
    'green_2': '#89AE8F',
    'green_3': '#638b66',
    'green_4': '#445F46',
    'gold_1': '#ffb04f',
    'gold_2': '#de9945',
    'gold_3': '#af7635',
    'gold_3b': '#e2c29d',
    'salmon_1': '#ff7371',
    'salmon_2': '#d6635f',
    'salmon_3': '#b65551',
    'purple_1': '#CA4A72',
    'purple_2': '#9D496D',
    'purple_3': '#772D4D',
    'crimson_1': '#AD134C',
    'lime_1': '#D4DD7F',
    'lime_2': '#9DAF7D',
    'lime_3': '#808D69',
    'brown_1': '#E4B49C',
    'brown_2': '#BF936F',
    'brown_3': '#9F7553',
    'teal_1': '#7FE1E1',
    'teal_2': '#5BC8C5',
    'teal_3': '#4E9A9C'
}

NEU5 = {
    'tan': '#CEA07E',
    'purple': '#8F6593',
    'blue': '#4F759B',
    'green': '#638b66',
    'red': '#B9314F'
}

NEU5L = {
    'black': '#676765',
    'tan': '#C7916B',
    'purple': '#A179A4',
    'blue': '#5C83AD',
    'green': '#709973',
    'red': '#D86E85'  
}

# Plotly Sheet Theme
def NORD_theme():
    plotly_NORD_theme = pio.templates["simple_white"]
    plotly_NORD_theme.layout.plot_bgcolor = "#f4f4f5" 
    plotly_NORD_theme.layout.paper_bgcolor = "#FFFFFF"
    plotly_NORD_theme.layout.xaxis.gridcolor = NORD['snow_3']
    plotly_NORD_theme.layout.yaxis.gridcolor = NORD['snow_3']
    return plotly_NORD_theme

def NORD_theme_dark():
    plotly_NORD_theme = pio.templates["plotly_dark"]
    plotly_NORD_theme.layout.plot_bgcolor = "rgb(25,25,25)"
    #plotly_NORD_theme.layout.paper_bgcolor = "#FFFFFF"
    plotly_NORD_theme.layout.xaxis.gridcolor = NORD['polar_3']
    plotly_NORD_theme.layout.yaxis.gridcolor = NORD['polar_3']
    return plotly_NORD_theme


def set_font_size(layout, font_size=16):
    layout['titlefont']['size'] = font_size + 4
    layout.legend['font']['size'] = font_size

    for ax in [item for item in layout if item.__contains__('xaxis')]:
        layout[ax].titlefont.size = font_size
        layout[ax].tickfont.size = font_size

    for ax in [item for item in layout if item.__contains__('yaxis')]:
        layout[ax].titlefont.size = font_size
        layout[ax].tickfont.size = font_size


class plt_size():
    def slim():
        layout = dict(
            width=1000,
            height=250
        )
        return layout
    def small():
        layout = dict(
            width=1000,
            height=450
        )
        return layout
    def standard():
        layout = dict(
            width=1000,
            height=600
        )
        return layout
    def medium():
        layout = dict(
            width=1000,
            height=800
        )
        return layout
    def tall():
        layout = dict(
            width=1000,
            height=1000
        )
        return layout

class plt_markup():
    def legend_bottom():
        layout = dict(
            legend=dict(
                xanchor='center',
                x=0.5,
                y=-0.1,
                orientation='h', 
            )
        )
        return layout

    def report_margins():
        layout = dict(
            title=None,
            margin=dict(l=20, r=20, t=20, b=0),
        )
        return layout

class plt_save():
    def svg(fig, filename):
        width_in_mm, width_default_px, dpi = 160, 1000, 800
        scale = (width_in_mm / 25.4) / (width_default_px / dpi)
        fig.write_image(f'{filename}.svg',scale=scale)
        return


import pandas as pd
import numpy as np
from mip import *

class Opt_Solver:
	def __init__(self, solver_package):
		self.m = Model(sense=MINIMIZE, solver_name=solver_package)
		self._optimise_timeout = 120 # timeout at 2 minutes, 120 seconds
		self._decision_vars = {}
		self._obj_func_components = None
		self._constr_lhs = None
		self._constr_rhs = None
        
	def add_input_variables(self, input_variables):
		input_variables = input_variables.sort_values('variable_id')
		# Update mip solver variable type column
		typeentries = list(input_variables['type'].unique())
		if 'binary' in typeentries:
			typeentries.remove('binary')
			input_variables['type'] = np.where(input_variables['type']=='binary',BINARY,input_variables['type'])
		if 'continuous' in typeentries:
			typeentries.remove('continuous')
			input_variables['type'] = np.where(input_variables['type']=='continuous',CONTINUOUS,input_variables['type'])
		if not len(typeentries) == 0:
			err = typeentries
			raise Exception(f"Unknown variable type {err} found in df")

		# Add variables to mip solver
		for idx, var_id in enumerate(input_variables['variable_id']):
			self._decision_vars[var_id] = self.m.add_var(lb=input_variables.loc[idx,'lower_bound'],
												ub=input_variables.loc[idx,'upper_bound'],
												var_type=input_variables.loc[idx,'type'],
												name=str(var_id))

	def add_objective_function(self, objective_costs):
		objective_costs = objective_costs.sort_values('variable_id')
		# Minimise violation costs
		obj_fnc = minimize(xsum(objective_costs[objective_costs['variable_id']==i]['cost'].values[0] * self._decision_vars[i] for i in
						  objective_costs['variable_id']))
		self.m.objective = obj_fnc   

	def add_constraints(self, lhs_df, rhs_df):
		rhs_df = rhs_df.sort_values('constraint_id')
		# Iterate through rhs dataframe to add constraint equations
		for idx in rhs_df['constraint_id']:
			element_list = []
			lhs_elements = lhs_df[lhs_df['constraint_id']==idx]
			for variable in lhs_elements['variable_id']:
				multiplier = lhs_elements[lhs_elements['variable_id']==variable]['coefficient'].values[0]
				element = self._decision_vars[variable] * multiplier
				element_list += [element]

			expression = xsum(element_list)
			inequality = rhs_df[rhs_df['constraint_id']==idx]['type'].values[0]
			rhs_value = rhs_df[rhs_df['constraint_id']==idx]['rhs'].values[0]

			if inequality == '<=':
				constraint = expression <= rhs_value
			elif inequality == '==':
				constraint = expression == rhs_value
			elif inequality == '>=':
				constraint = expression >= rhs_value
			else:
				raise Exception("Error with format of inequality type")
			self.m.add_constr(constraint, name=str(idx))
                
	def add_constraints_dynamic(self, lhs_df, rhsD_df):
        # Add dynamic rhs constraints
		if not rhsD_df.empty:
			rhsD_df = rhsD_df.sort_values('constraint_id')
			for idx in rhsD_df['constraint_id']:
				# Find RHS variable
				inequality = rhsD_df[rhsD_df['constraint_id']==idx]['type'].values[0]
				rhs_var_id = rhsD_df.loc[rhsD_df['constraint_id']==idx, 'rhs_variable_id'].values[0]
				rhs_value = self._decision_vars[rhs_var_id]
                
				lhs_elements = lhs_df[lhs_df['constraint_id']==idx]
				# if rhsD_df[rhsD_df['constraint_id']==idx]['interval'].values[0] != None:
				# 	# Muliply LHS variable by coefficient.
				# 	lhs_var_id = lhs_elements['variable_id'].values[0]
				# 	multiplier = lhs_elements['coefficient'].values[0]
				# 	expression = self._decision_vars[lhs_var_id] * multiplier
				# else:
				# The constraint is a summation on lhs to set variable on rhs (used for ppa_lgc_sum_1)
				element_list = []
				for variable in lhs_elements['variable_id']:
					multiplier = lhs_elements[lhs_elements['variable_id']==variable]['coefficient'].values[0]
					element = self._decision_vars[variable] * multiplier
					element_list += [element]
				expression = xsum(element_list)
                    
				# if inequality == '==':
				# 	# Prevent error with optimisation trying to set 0 == variable
				# 	constraint = rhs_value == expression
				# else:
				# 	raise NotImplementedError("Dynamic RHS constraints does not yet allow >= or <= constraints")


				if inequality == '<=':
					constraint = expression <= rhs_value
				elif inequality == '==':
					constraint = rhs_value == expression # Prevent error with optimisation trying to set 0 == variable
				elif inequality == '>=':
					constraint = expression >= rhs_value
				else:
					raise Exception("Error with format of inequality type")

				self.m.add_constr(constraint, name=str(idx))
		
	def add_constraints_bigM(self, lhs_df, bigM_df):
		# Apply Big M constraints
		if not bigM_df.empty:
			bigM_df = bigM_df.sort_values('constraint_id')
			for idx in bigM_df['constraint_id']:
				# Define expression as LHS variable
				lhs_var_idx = bigM_df.loc[bigM_df['constraint_id']==idx, 'lhs_variable_id'].values[0]
				lhs_exp = self._decision_vars[lhs_var_idx]
                
				# Define rhs expression and other fields
				inequality = bigM_df.loc[bigM_df['constraint_id']==idx, 'constr_type'].values[0]
				mode = bigM_df.loc[bigM_df['constraint_id']==idx, 'mode'].values[0]
				Mval = bigM_df.loc[bigM_df['constraint_id']==idx, 'bigM'].values[0]
				rhs_var_idx = bigM_df.loc[bigM_df['constraint_id']==idx, 'rhs_variable_id'].values[0]
				rhs_bvar = self._decision_vars[rhs_var_idx]

				if mode == 'normal':
					rhs_exp = Mval * rhs_bvar
				elif mode == 'inverse':
					rhs_exp = Mval * (1 - rhs_bvar)
				else:
					raise Exception("BigM mode is invalid")
                
				if inequality == '<=':
					constraint = lhs_exp <= rhs_exp
				elif inequality == '==':
					constraint = lhs_exp == rhs_exp
				elif inequality == '>=':
					constraint = lhs_exp >= rhs_exp
				else:
					raise Exception("Error with format of inequality type")
				self.m.add_constr(constraint, name=str(idx))
				
	def add_sos_type_2_ARCHIVE(self, sos_2_df):

		#self._decision_vars[lhs_var_idx]

		for interval in sos_2_df['interval'].unique()[0:1]:

			sos_entry = sos_2_df[sos_2_df['interval']==interval]

			#sos_entry['sos_varid'] = sos_entry['sos_varid'].apply(lambda x: self._decision_vars[x])
			sos_entry['sos_varid'].update(pd.Series(self._decision_vars))

			tupleobj = [(sos_entry.loc[row,'sos_varid'], sos_entry.loc[row,'sos_val']) for row in sos_entry.index]

			self.m.add_sos(tupleobj,2)

		# pass in tuple with weight var name and x-point


		#self.m.add_sos(,2)
		return

	def add_sos_type_2(self, sos_name, weights_df, x_samples, y_samples, xvar_df, yvar_df):

		for interval in weights_df['interval'].unique():
			weights_ids = weights_df.loc[weights_df['interval']==interval, 'variable_id']
			weights_ids = weights_ids.sort_values().to_list()
			weights = [self._decision_vars[k] for k in weights_ids]

			xvar_id = xvar_df.loc[xvar_df['interval']==interval, 'variable_id']
			xvar = self._decision_vars[int(xvar_id)]

			yvar_id = yvar_df.loc[yvar_df['interval']==interval, 'variable_id']
			yvar = self._decision_vars[int(yvar_id)]

			# Weights
			self.m.add_sos([(weights[i], x_samples[i]) for i in range(len(x_samples))], 2)
		
			# Constraint for sum of weights 
			sum_of_weights = xsum(weights) == 1
			self.m.add_constr(sum_of_weights, name=sos_name+"_wsum")
			
			# Constraint for linking x_samples
			x_link = xsum([weights[i]*x_samples[i] for i in range(len(x_samples))]) == xvar
			self.m.add_constr(x_link, name=sos_name+"_xlink")

			# Constraint for linking y_samples
			y_link = xsum([weights[i]*y_samples[i] for i in range(len(y_samples))]) == yvar
			self.m.add_constr(y_link, name=sos_name+"_ylink")


		return

	def optimise(self):
		status = self.m.optimize(max_seconds=self._optimise_timeout)
		print(f"OPTIMISATION COMPLETE, Obj Value: {self.m.objective_value}")
		if status != OptimizationStatus.OPTIMAL:
			print("%%%%%%%%%%%%%%%%%%%%% INFEASIBLE MODEL %%%%%%%%%%%%%%%%%%%%%")
			print(f"STATUS: {OptimizationStatus}")
			con_index, newsvr = self.isolate_problem()
			print('Couldn\'t find an optimal solution, but removing con {} fixed INFEASIBLITY'.format(con_index))
			print("INFEASIBLE")#raise ValueError('Linear program infeasible')
			return newsvr, con_index
		else:
			return self.m, None
        

	def isolate_problem(self):
		"""Credit: N.Gorman (nempy)
		"""
		base_prob = self.m
		cons = []
		test_prob = base_prob.copy()
		for con in [con.name for con in base_prob.constrs]:
			[test_prob.remove(c) for c in test_prob.constrs if c.name == con]
			status = test_prob.optimize(max_seconds=self._optimise_timeout)
			cons.append(con)
			if status == OptimizationStatus.OPTIMAL:
				return cons, test_prob
		return []


	def get_decision_variable_vals(self, solver, var_df):
		result = var_df['variable_id'].apply(lambda x: solver.var_by_name(str(x)).x)
		var_df.insert(5,'value',result)
		return var_df
        
	def get_constraint_slack(self, solver, rhs_df):
		result = rhs_df['constraint_id'].apply(lambda x: solver.constr_by_name(str(x)).slack)
		rhs_df.insert(5,'slack',result)
		return rhs_df





import yaml
import json
from nemglo.planning.data_fetch import Market
from nemglo.components.renewables import Generator
from nemglo.components.electrolyser import Electrolyser
from nemglo.components.emissions import Emissions
from nemglo.planning.planner import Plan 
from nemglo.backend import input_validation as inv
from nemglo.defaults import *
from types import SimpleNamespace
from datetime import datetime as dt
import logging

logger = logging.getLogger(__name__)

def convert_dt_fmt(string):
    ret_str = dt.strptime(string, '%Y-%m-%d')
    ret_str = dt.strftime(ret_str, '%Y/%m/%d')
    return ret_str

def convert_timestamp(timestamps):
    result = []
    for time in timestamps:
        result.append(dt.strptime(str(time), '%Y-%m-%d %H:%M:%S').strftime('%Y-%m-%d %H:%M:%S'))
    return result


def get_market_data(conf):
    """Processes market data API call.

    Parameters
    ----------
    conf : dict
        Input parameters to process. As defined in NEMGLO-API documentation.

    Returns
    -------
    json
        Returned Fields: "availgens", "availgens_default_cap", "availgens_tech", "time", "timestamps", "prices"
    """
    logger.info('=== Commence get_market_data() ===')
    inv.validate_get_market_data(conf)
    c_md = SimpleNamespace(**conf['market_data'])

    # Nemosis Data -> Price extraction

    data = Market(local_cache=DATA_CACHE.FILEPATH,
                  intlength=int(c_md.dispatch_interval_length),
                  region=c_md.region,
                  start_date=convert_dt_fmt(c_md.start_date) + " " + [c_md.start_time if hasattr(c_md, "start_time") else "00:00"][0],
                  end_date=convert_dt_fmt(c_md.end_date) + " " + [c_md.end_time if hasattr(c_md, "end_time") else "00:00"][0],
                  )
    data._download_geninfo(filter_regions=[c_md.region])
    tracedata, tracedata['time'], tracedata['prices'], tracedata['vre'] = {}, {}, {}, {}

    # Nemed Data -> Emissions extraction
    if 'emissions_data' in conf:
        c_em = SimpleNamespace(**conf['emissions_data'])
        data._emissions_type = c_em.emissions_type
        print(data._emissions_type)
        try:
            co2_df = data.get_emissions_intensity()
        except Exception as e:
            print(e)


    prc_df = data.get_prices()
    tracedata['time'] = prc_df['Time']
    tracedata['prices'] = prc_df['Prices']
    result = {}
    result['availgens'] = data._info['DUID'].to_list()
    result['availgens_default_cap'] = data._info['Reg Cap (MW)'].to_list()
    result['availgens_tech'] = data._info['Fuel Source - Descriptor'].to_list()
    result['time'] = convert_timestamp(tracedata['time'])
    result['timestamps'] = json.loads(tracedata['time'].dt.tz_localize('Australia/Sydney').to_json(orient='records'))
    result['prices'] = tracedata['prices'].to_list()
    if ('emissions_data' in conf):
        result['emissions'] = co2_df['Intensity_Index'].to_list()
    logger.info('=== End get_market_data() ===')
    return json.dumps(result)


def get_generator_data(conf):
    """Processes generator data API call. (Single DUID at a time)

    Parameters
    ----------
    conf : dict
        Input parameters to process. As defined in NEMGLO-API documentation.

    Returns
    -------
    json
        Returned Fields: "generator", "time", "timestamps", "cf_trace", "mw_trace", "prices", "cost_trace",
        "metric_cf_avg"
    """
    logger.info('=== Commence get_generator_data() ===')
    inv.validate_get_generator_data(conf)
    c_md = SimpleNamespace(**conf['market_data'])
    c_ppa = SimpleNamespace(**conf['ppa'])
    data = Market(local_cache=DATA_CACHE.FILEPATH,
                  intlength=int(c_md.dispatch_interval_length),
                  region=c_md.region,
                  start_date=convert_dt_fmt(c_md.start_date) + " " + \
                    [c_md.start_time if hasattr(c_md, "start_time") else "00:00"][0],
                  end_date=convert_dt_fmt(c_md.end_date) + " " + \
                    [c_md.end_time if hasattr(c_md, "end_time") else "00:00"][0],
                  duid_1=c_ppa.duid)
    tracedata, tracedata['time'], tracedata['vre'] = {}, {}, {}
    price = data.get_prices() # Run to load price traces attr to Market object
    vre = data.get_generation()
    print(c_ppa.floor_price)
    cost = data.estimate_ppa_costs(vre_capacity=float(c_ppa.capacity),
                                   strike_price=float(c_ppa.strike_price),
                                   floor_price=[None if c_ppa.floor_price == None else \
                                                None if c_ppa.floor_price == "None" else \
                                                float(c_ppa.floor_price)][0])
    tracedata['time'] = vre['Time']
    tracedata['vre'] = vre[data._duid_1] 
    result = {}
    result['generator'] = [tracedata['vre'].name]
    result['time'] = convert_timestamp(tracedata['time'])
    result['timestamps'] = json.loads(tracedata['time'].dt.tz_localize('Australia/Sydney').to_json(orient='records'))
    result['cf_trace'] = json.loads(tracedata['vre'].round(4).to_json(orient='records'))
    result['mw_trace'] = json.loads(tracedata['vre'].mul(c_ppa.capacity).round(2).to_json(orient='records'))
    result['prices'] = json.loads(price['Prices'].round(2).to_json(orient='records'))
    result['cost_trace'] = json.loads(cost['Cost'].round(2).to_json(orient='records'))
    result['metric_cf_avg'] = tracedata['vre'].mean().round(4)
    logger.info('=== End get_generator_data() ===')
    return json.dumps(result)


def get_operation(conf):
    # Process Inputs
    logger.info('=== Commence get_operation() ===')
    # inv.validate_get_operation(conf)
    c_md = SimpleNamespace(**conf['market_data'])
    c_el = SimpleNamespace(**conf['electrolyser_load'])
    if 'ppa_1' in conf:
        c_ppa1 = SimpleNamespace(**conf['ppa_1'])
    else:
        c_ppa1 = None
    if 'ppa_2' in conf:
        c_ppa2 = SimpleNamespace(**conf['ppa_2'])
    else:
        c_ppa2 = None
    if 'rec' in conf:
        c_rec = SimpleNamespace(**conf['rec'])
    else:
        c_rec = None
    if 'emissions' in conf:
        c_em = SimpleNamespace(**conf['emissions'])
    else:
        c_em = None

    # Setup environment
    tracedata, tracedata['time'], tracedata['prices'], tracedata['vre'] = {}, {}, {}, {}
    p2g = Plan(identifier="p2g")

    # Market Data
    data = Market(local_cache=DATA_CACHE.FILEPATH,
                  intlength=int(c_md.dispatch_interval_length),
                  region=c_md.region,
                  start_date=convert_dt_fmt(c_md.start_date) + " " + [c_md.start_time if hasattr(c_md, "start_time") else "00:00"][0],
                  end_date=convert_dt_fmt(c_md.end_date) + " " + [c_md.end_time if hasattr(c_md, "end_time") else "00:00"][0],
                  duid_1=[c_ppa1.duid if c_ppa1 != None else None][0],
                  duid_2=[c_ppa2.duid if c_ppa2 != None else None][0])

    prc_df = data.get_prices()
    tracedata['time'] = prc_df['Time']
    tracedata['prices'] = prc_df['Prices']
    p2g.load_market_prices(prc_df)
    tracedata['vre'] = data.get_generation()  
    logger.info("+ Market Data Retrieved")

    # Electrolyser Data
    h2e = Electrolyser(p2g, identifier="h2e")
    h2e.load_h2_parameters_preset(capacity = float(c_el.capacity),
                                maxload = float(c_el.capacity),
                                minload = float(c_el.min_stable_load),
                                offload = 0.0,
                                electrolyser_type = str(c_el.technology_type),
                                sec_profile = str(c_el.sec_profile),
                                h2_price_kg = float(c_el.h2_price))
    h2e.add_electrolyser_operation()
    logger.info("+ Electrolyser Configured")

    # PPA Data
    if 'ppa_1' in conf:
        gen_1 = Generator(p2g, identifier='gen_1')
        gen_1.load_vre_parameters(duid = str(c_ppa1.duid),
                                  capacity = float(c_ppa1.capacity),
                                  trace = tracedata['vre'][['Time', c_ppa1.duid]],
                                  ppa_strike = float(c_ppa1.strike_price),
                                  ppa_floor = [float(c_ppa1.floor_price) if hasattr(c_ppa1, "floor_price") else None][0])
        logger.info("+ PPA1 Parameters loaded")
        gen_1.add_ppa_contract()
        logger.info("+ PPA1 Configured")
    if 'ppa_2' in conf:
        gen_2 = Generator(p2g, identifier='gen_2')
        gen_2.load_vre_parameters(duid = str(c_ppa2.duid),
                                  capacity = float(c_ppa2.capacity),
                                  trace = tracedata['vre'][['Time', c_ppa2.duid]],
                                  ppa_strike = float(c_ppa2.strike_price),
                                  ppa_floor = [float(c_ppa2.floor_price) if hasattr(c_ppa2, "floor_price") else None][0])
        logger.info("+ PPA2 Parameters loaded")
        gen_2.add_ppa_contract()
        logger.info("+ PPA2 Configured")

    # Certificate Data
    if 'rec' in conf:
        # print('REC params: {}, {}'.format(c_rec.constraint, type(c_rec.rec_price)))
        if hasattr(c_rec, "rec_price"):
            rec_price = float(c_rec.rec_price)
        else:
            rec_price = 0.0

        if c_rec.constraint == "interval":
            p2g.autoadd_rec_price_on_interval(rec_price=rec_price, allow_buying=c_rec.allow_buying, allow_selling=c_rec.allow_selling)
            logger.info("+ Applied REC constraint on interval")
        elif c_rec.constraint == "total":
            p2g.autoadd_rec_price_on_total(rec_price=rec_price, allow_buying=c_rec.allow_buying, allow_selling=c_rec.allow_selling)
            logger.info("+ Applied REC constraint on total")

    logger.info("# CONFIG")
    logger.info(conf)

    # Emissions Data
    if 'emissions' in conf:
        # Fetch emissions data via NEMED
        data._emissions_type = c_em.emissions_type
        co2_df = data.get_emissions_intensity()
        logger.info("-- Downloaded grid emissions data --")
        em = Emissions(p2g, identifier="em")
        if hasattr(c_em, "co2_price"):
            # Shadow carbon price applied
            em.load_emissions(co2_df, float(c_em.co2_price))
            em.add_emissions()
            em._price_emissions()
            logger.info("+ Applied Shadow Price on Grid Emissions Intensity")

        elif hasattr(c_em, "co2_constraint"):
            # or Carbon constraint applied
            em.load_emissions(co2_df, 0.0)
            em.add_emissions()
            em._set_emissions_limit(float(c_em.co2_constraint))
            logger.info("+ Applied Carbon Intensity of H2 Constraint considering Grid Emissions Intensity")

    # Optimise
    logger.info(">> Attempt Optimisation")
    p2g.optimise()
    logger.info("+ Optimisation Success")
    
    costs = p2g.get_costs(exclude_shadow=True)
    if ('rec' in conf):
        if (c_rec.constraint == "interval"):
            rec_df = p2g.get_rec_on_interval(as_mwh=False)
    
    # Results
    result = {}
    result['time'] = convert_timestamp(tracedata['time'])
    result['prices'] = tracedata['prices'].to_list()
    logger.info("+ Results: Prices")

    if ('ppa_1' in conf):
        result['ppa1'] = json.loads(tracedata['vre'][c_ppa1.duid].to_json(orient='split'))
        result['size_vre1'] = c_ppa1.capacity
        result['cost_vre1'] = costs['gen_1-ppa_cfd'].to_list()
    if ('ppa_2' in conf):
        result['ppa2'] = json.loads(tracedata['vre'][c_ppa2.duid].to_json(orient='split'))
        result['size_vre2'] = c_ppa2.capacity
        result['cost_vre2'] = costs['gen_2-ppa_cfd'].to_list()
    if ('ppa_1' in conf) & ('ppa_2' in conf):
        tracedata['combined_vre'] = c_ppa1.capacity * tracedata['vre'][c_ppa1.duid] + \
                                    c_ppa2.capacity * tracedata['vre'][c_ppa2.duid]
        comb_vre = json.loads(tracedata['combined_vre'].to_json(orient='split'))
        result['combined_vre'] = comb_vre['data']
    logger.info("+ Results: PPAs")

    # If REC is setup on interval and either of buy or sell (this feature is enabled)
    if ('rec' in conf):
        if (c_rec.constraint == "interval"):
            if (c_rec.allow_buying) & (not c_rec.allow_selling):
                comb_rec = costs['p2g-mkt_rec_buy']
                if (c_rec.constraint == "interval"):
                    rec_df['mkt'] = rec_df['Market RECs bought']
                else:
                    rec_df['mkt'] = 0
            elif (not c_rec.allow_buying) & (c_rec.allow_selling):
                comb_rec = costs['p2g-mkt_rec_sell']
                if (c_rec.constraint == "interval"):
                    rec_df['mkt'] = -1 * rec_df['Market RECs sold']
                else:
                    rec_df['mkt'] = 0
            elif (c_rec.allow_buying) & (c_rec.allow_selling):
                comb_rec = costs['p2g-mkt_rec_buy'] + costs['p2g-mkt_rec_sell']
                if (c_rec.constraint == "interval"):
                    rec_df['mkt'] = rec_df['Market RECs bought'] - rec_df['Market RECs sold']
                else:
                    rec_df['mkt'] = 0
        
            # Cost results as long as price exists
            if hasattr(c_rec, "rec_price") & (c_rec.allow_buying | c_rec.allow_selling):
                if c_rec.rec_price != None:
                    result['cost_rec'] = json.loads(comb_rec.to_json(orient='split'))['data']
                    result['rec'] = rec_df['mkt'].to_list()
        logger.info("+ Results: RECs")


    if ('emissions' in conf):
        # Return grid emissions intensity
        result['grid_emissions'] = co2_df['Intensity_Index'].to_list()
        result['grid_load'] = p2g._format_out_vars_timeseries('em-grid_load')['value'].to_list()
        result['vre_load'] = p2g._format_out_vars_timeseries('em-vre_proportion')['value'].to_list()
        result['impact_emissions'] = p2g._format_out_vars_timeseries('em-impact_emissions')['value'].to_list()
        if hasattr(c_em, "co2_price"):
            # Return cost component for emissions
            costs = p2g.get_costs(exclude_shadow=False)
            result['cost_emissions'] = costs['em-impact_emissions'].to_list()

    tracedata['optimised_load'] = p2g.get_load()
    tracedata['optimised_load']['time'] = tracedata['optimised_load']['time'].dt.tz_localize('Australia/Sydney')
    opt_load = json.loads(tracedata['optimised_load'].to_json(orient='split'))
    result['optimised_load'] = [x[1] for x in opt_load['data']]
    result['timestamps'] = [x[2] for x in opt_load['data']]
    logger.info("+ Results: Optimised Load")

    result['cost_total'] = costs['total_cost'].to_list()
    result['cost_h2'] = costs['h2e-h2_produced'].to_list()
    result['cost_energy'] = costs['h2e-mw_load'].to_list()
    logger.info("+ Results: Costs")

    logger.info('=== End get_operation() ===')
    return json.dumps(result)



import pandas as pd
import numpy as np
from nemglo.backend.optimiser_formatters import *
from nemglo.backend import input_validation as inv
from nemglo.planning.planner import Plan

class Emissions:
    """Object to store the user input parameters of Emissions, check and validate inputs,
    then perform loading actions to planner object.
    """
    def __init__(self, system_plan, identifier):
        # Error check on parsed arguments
        assert isinstance(system_plan, Plan), "Generator Argument: 'system_plan' must be nemglo.planner.Plan object"
        assert isinstance(identifier, str), "Generator Argument: 'identifier' must be a str"
        inv.validate_unique_id(self.__class__, system_plan, identifier)

        # Link object to Plan
        self._system_plan = system_plan
        cname = self.__class__.__name__
        self._system_plan._components.update({cname: [self]})
        self._id = identifier

        # Emissions Characteristics
        self._trace = None # tCO2-e
        self._shadow_price = None # $/tCO2-e

    
    def load_emissions(self, trace, shadow_price):
        # Validate Inputs
        assert isinstance(shadow_price, float), "Emissions Argument: 'shadow_price' must be a float"
        self._validate_trace(trace)

        # Store Values
        self._trace = trace
        self._shadow_price = shadow_price


    def _validate_trace(self, trace):
        schema = inv.DataFrameSchema(name="trace", primary_keys=['Time', 'Intensity_Index'])
        schema.add_column(inv.SeriesSchema(name='Time', data_type=np.dtype('datetime64[ns]'), no_duplicates=True, \
            ascending_order=True, minimum=self._system_plan._timeseries[0], maximum=self._system_plan._timeseries[-1]))
        schema.add_column(inv.SeriesSchema(name='Intensity_Index', data_type=np.float64, must_be_real_number=True, \
            not_negative=True))
        schema.validate(trace)


    def add_emissions(self):
        planner = self._system_plan
        vre_prop_name = self._id + '-vre_proportion'
        co2_name = self._id + '-impact_emissions'
        grid_name = self._id + '-grid_load'
        load_name = planner._components['Electrolyser'][0]._id + '-mw_load'
        load_cap = planner._components['Electrolyser'][0]._capacity

        # > If there's at least one vre generator...
        if 'Generator' in planner._components:
            # Create `vre_proportion` variable equal to vre_avail but capped at load capacity. Priced in opt to minimise var.
            create_timeseries_vars(planner, var_name=vre_prop_name, lb=0, ub=np.inf)
            # vre_prop will not exceed max load
            create_constr_rhs_on_interval(planner, constr_name=vre_prop_name+"_maxload", constr_type="<=", rhs_value=load_cap) 
            create_constr_lhs_on_interval(planner, constr_name=vre_prop_name+"_maxload",
                constr_rhs=planner._constr_rhs[vre_prop_name+"_maxload"], coefficient_map={vre_prop_name: 1})
            # vre_prop will not exceed combined vre avails
            var_ids = planner._var[vre_prop_name][['interval', 'variable_id']]
            create_constr_rhs_on_interval_dynamicvar(planner, constr_name=vre_prop_name+"_avails",
                constr_type='>=', rhs_var_id_series=var_ids)
            coeffs = {}
            for gen in planner._components['Generator']:
                name_ppa_rec = gen._id + '-vre_avail'
                coeffs.update({name_ppa_rec: 1})
            create_constr_lhs_on_interval(planner, constr_name=vre_prop_name+"_avails",
                constr_rhs=planner._constr_rhs_dynamic[vre_prop_name+"_avails"], coefficient_map=coeffs)
            # Incentivise var to max out to upper limit
            create_objective_cost(planner, var_name=vre_prop_name, decision_var_series=planner._var[vre_prop_name], cost=-1)
        else:
            create_timeseries_vars(planner, var_name=vre_prop_name, lb=0, ub=0)

        # Create `grid_load` variable which is mw_load of electrolyser less RE propotion
        create_timeseries_vars(planner, var_name=grid_name, lb=0, ub=np.inf)
        
        # Set `impact emissions` variable
        var_ids = planner._var[grid_name][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=grid_name,
            constr_type='==', rhs_var_id_series=var_ids)

        coeffs = {load_name: 1, vre_prop_name: -1}

        create_constr_lhs_on_interval(planner, constr_name=grid_name, constr_rhs=planner._constr_rhs_dynamic[grid_name],
                                        coefficient_map=coeffs)


        # Create `impact emissions` variable
        create_timeseries_vars(planner, var_name=co2_name, lb=0, ub=np.inf)
        # Set `impact emissions` variable
        var_ids = planner._var[co2_name][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=co2_name,
            constr_type='==', rhs_var_id_series=var_ids)
        # Set `impact emissions` volume
        coeffs = self._trace['Intensity_Index'].mul(planner._intlength / 60).to_list()
        create_constr_lhs_on_interval_dynamic(planner, constr_name=co2_name,
            constr_rhs=planner._constr_rhs_dynamic[co2_name],
            decision_vars=planner._var[grid_name],
            coefficient_list=coeffs)




    def _price_emissions(self):
        planner = self._system_plan
        co2_name = self._id + '-impact_emissions'

        # Price `impact emissions` in optimiser
        series = planner._var[co2_name]
        create_objective_cost(planner, var_name=co2_name,
                    decision_var_series=series, cost=self._shadow_price)


    def _set_emissions_limit(self, limit_value): #limit_value in tCo2/tH2
        planner = self._system_plan
        co2_name = self._id + '-impact_emissions'
        co2tot_name = self._id + '-total_emissions'
        const_nm = self._id + '-em_limit'
        prodsum_name = planner._components['Electrolyser'][0]._id + '-h2_produced_sum'

        # Create variable for total emissions (kg) over simulation
        create_var(planner, var_name=co2tot_name, lb=0, ub=np.inf)
        rhs_var_id = planner._var[co2tot_name]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(planner, constr_name=co2tot_name, constr_type="==", rhs_var_id=rhs_var_id)
        create_constr_lhs_on_interval_fixed_rhs(planner, constr_name=co2tot_name,
                                                    constr_rhs=planner._constr_rhs_dynamic[co2tot_name],
                                                    decision_vars=planner._var, coefficient_map={co2_name: 1000})

        # Create const for sum of emissions <= limit value
        # limit_value (tCo2/tH2) * h2 production >= total_emissions
        rhs_var_id = planner._var[co2tot_name]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(planner, constr_name=const_nm, constr_type=">=", rhs_var_id=rhs_var_id)
        create_constr_lhs_on_interval(planner, constr_name=const_nm, constr_rhs=planner._constr_rhs_dynamic[const_nm],\
            coefficient_map={prodsum_name: limit_value})

        # sum emissions <= 3 * h2 production
        # tCO2/MWh
        # 100 tCO2 for 1000 kg
        # kg CO2 per kg H2

        


import pandas as pd
import numpy as np
from datetime import timedelta
from nemglo.planning.planner import Plan
from nemglo.backend import input_validation as inv
from nemglo.backend.optimiser_formatters import *
from nemglo.defaults import *

class Electrolyser:
    """Object to store the user input parameters of electrolyser, check and validate inputs, then perform loading
    actions to planner object.

    Parameters
    ----------
    system_plan : nemglo.planner.Plan
        The system plan object for which the Electrolyser should be linked to.
        
    identifier : str
        A unique identifier to refer to the object as. For example, 'H2E'.

    """
    def __init__(self, system_plan, identifier):
        # Error check on parsed arguments
        assert isinstance(system_plan, Plan), "Electrolyser Argument: 'system_plan' must be nemglo.planner.Plan object"
        assert isinstance(identifier, str), "Electrolyser Argument: 'identifier' must be a str"
        inv.validate_unique_id(self.__class__, system_plan, identifier)

        # Link object to Plan
        self._system_plan = system_plan
        cname = self.__class__.__name__
        self._system_plan._components.update({cname: [self]})
        self._id = identifier

        # Generic Load Characteristics
        self._capacity = None
        self._maxload = None
        self._minload = None
        self._offload = None

        # (Electricity -> Hydrogen Production) Conversion - Specific Energy Consumption
        self._type = None
        self._profile = 'fixed'
        self._sec_nominal = None  # kWh/kg
        self._sec_conversion = None # set default at 100%
        self._sec_system = None  # Multiplication of _nominal and _conversion factor

        # Internally defined parameters
        self._sec_variable_points = None # pd.Dataframe mapping x-y points

        # Hydrogen only model - production benefit price
        self._h2_price_kg = None # $/kg

        # Hydrogen storage
        self._storage_max = None # Max storage in kg H2
        self._storage_initial = None
        self._storage_final = None
        self._storage_drain = None


    # ========================= Input & User Functions =========================
    def load_h2_parameters_preset(self, capacity, maxload, minload, offload, electrolyser_type, sec_profile, 
        h2_price_kg=None):
        """Load Electrolyser parameters to define the object.

        Parameters
        ----------
        capacity : float
            The desired rated capacity of the Electrolyser, in MW.
        maxload : float
            The maximum load operation of the Electrolyser, in MW. At present, NEMGLO assumes capacity parameter.
            `maxload` has null effect on the optimiser.
        minload : float
            The minimum stable loading (MSL) of the Electrolyser, in MW.
        offload : float
            The off state of the Electrolyser, in MW. In other words, absolute minimum value.
        electrolyser_type : str
            Either set as 'PEM' or 'AE'. This determines which set of predefined electrolyser characteristics are used.
        sec_profile : str
            Either set as 'fixed' or 'variable'. This determines the behaviour of the conversion function from
            electricity, as MW, to hydrogen, as kg.
        h2_price_kg : float, optional
            The sale price or production benefit received for the volume of hydrogen produced, by default None.
        """
        # Validate Inputs
        inv.validate_positive_float(capacity, 'capacity', type(self).__name__)
        inv.validate_positive_float(maxload, 'maxload', type(self).__name__)
        inv.validate_positive_float(minload, 'minload', type(self).__name__)
        inv.validate_positive_float(offload, 'offload', type(self).__name__)
        assert isinstance(electrolyser_type, str), "Electrolyser Argument: 'electrolyser_type' must be a str"
        assert isinstance(sec_profile, str), "Electrolyser Argument: 'sec_profile' must be a str"
        assert isinstance(h2_price_kg, (float, type(None))), "Electrolyser Argument: 'h2_price_kg' must be a \
            (float, or None)"

        assert electrolyser_type in ['PEM','AE'], \
            "Electrolyser Argument: 'electrolyser_type' must be one of ['PEM','AE']"
        assert sec_profile in ['variable','fixed'], \
            "Electrolyser Argument: 'sec_profile' must be one of ['variable','fixed']"

        # Store Values
        self._capacity = capacity
        self._maxload = maxload
        self._minload = minload
        self._offload = offload
        self._type = electrolyser_type
        self._profile = sec_profile
        self._h2_price_kg = h2_price_kg

        if self._type == "PEM":
            self._sec_nominal = SEC_NOMINAL_PEM
            self._sec_conversion = SEC_CONVERSION_PEM
            self._sec_variable_points = pd.DataFrame(SEC_PROFILE_PEM)
        elif self._type == "AE":
            self._sec_nominal = SEC_NOMINAL_AE
            self._sec_conversion = SEC_CONVERSION_AE
            self._sec_variable_points = pd.DataFrame(SEC_PROFILE_AE)
        
        self._sec_system = self._sec_nominal * self._sec_conversion


    def load_h2_parameters_custom(self, capacity, maxload, minload, offload, sec_profile, sec_nominal, sec_conversion,\
        sec_datapoints, h2_price_kg=None):
        raise NotImplementedError("Custom parameters are not permitted in this version of NEMGLO")


    def load_h2_storage_parameters(self):
        raise NotImplementedError("Storage parameters are not permitted in this version of NEMGLO")


    def add_electrolyser_operation(self):

        # Verify input parameters exist
        assert (self._maxload is not None) and (self._offload is not None), \
            "Missing Electrolyser Inputs. Be sure to call `load_h2_parameters` before `add_`"

        # Define MW Load variable
        inv.validate_existing(self._system_plan, self._id+'-mw_load', as_var=True, as_objective_cost=True)
        self._set_mw_load()

        # Define MW Load Sum
        inv.validate_existing(self._system_plan, self._id+'-mw_load_sum', as_var=True, as_constr_rhs_dynamic=True, \
            as_constr_lhs=True)
        self._set_mw_load_sum()

        # Fixed or Variable SEC
        if self._profile == 'fixed':
            inv.validate_existing(self._system_plan, self._id+'-h2_produced', as_var=True, as_constr_rhs_dynamic=True, \
            as_constr_lhs=True)
            self._set_h2_production_fixed()
        elif self._profile == 'variable':
            inv.validate_existing(self._system_plan, self._id+'-h2_produced', as_var=True)
            inv.validate_existing(self._system_plan, self._id+'-sec', as_sos_2=True)
            self._set_h2_production_variable()
        
        # Define H2 Production Sum
        inv.validate_existing(self._system_plan, self._id+'h2_produced_sum', as_var=True, as_constr_rhs_dynamic=True, \
            as_constr_lhs=True)
        self._set_h2_production_sum()

        # Production benefit price
        if self._h2_price_kg:
            inv.validate_existing(self._system_plan, self._id+'-h2_produced', as_objective_cost=True)
            self._price_h2_production()

        if self._minload > 0:
            self._set_h2_msl_limit()


    def remove_electrolyser_operation(self):
        """Removes all assosciations of the Electrolyser object from the Plan object.
        """
        remove_element(self._system_plan, self._id, all_assoc=True)


    # ==================== Electrolyser Operation Functions ====================
    def _set_mw_load(self):
        """Defines Electrolyser MW load variables and prices variables according to input market prices in Plan object.
        """
        planner = self._system_plan
        name_load = self._id + '-mw_load'

        # Create Load variables
        create_timeseries_vars(planner, var_name=name_load, lb=self._offload, ub=self._maxload)

        # Extract planner module prices
        price_series = pd.DataFrame({'interval': range(planner._n), 'cost': [planner._prices[i] * \
            (planner._intlength/60) for i in range(0, len(planner._prices))]})
        
        # Price the variables in optimisation cost function
        create_objective_cost(planner, var_name=name_load, decision_var_series=planner._var[name_load], \
            cost=price_series)


    def _set_mw_load_sum(self):
        """Add a variable summating all interval data for the Electrolyser mw_load variables.
        """
        planner = self._system_plan
        name_load_sum = self._id + '-mw_load_sum'
        name_load = self._id + '-mw_load'

        # New variable for summation of load series
        create_var(planner, var_name=name_load_sum, lb=0.0, ub=np.inf)
        
        # New constr with variable as RHS
        rhs_var_id = planner._var[name_load_sum]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(planner, constr_name=name_load_sum, constr_type='==', rhs_var_id=rhs_var_id)

        # New lhs summation of constr
        create_constr_lhs_on_interval_fixed_rhs(planner, constr_name=name_load_sum,
                                                    constr_rhs=planner._constr_rhs_dynamic[name_load_sum],
                                                    decision_vars=planner._var, coefficient_map={name_load: 1})


    def _set_h2_production_fixed(self):
        """
        Creates a timeseries of Hydrogen Production variables for the optimiser determined by a fixed (linear) specific
        energy consumption (kWh/kg). Each variable is set by the equation of:
        h2_production [kg] = (1 / sec_system) [kg/kWh] * 10^3 [kWh/MWh] * h2e_load [MW] * \
            (Interval Length [min] / 60) [h]
        """
        planner = self._system_plan
        product_name = self._id + '-h2_produced'
        load_name = self._id + '-mw_load'

        # H2 production variable in kg
        create_timeseries_vars(planner, var_name=product_name, lb=0.0, ub=np.inf)
        
        # H2 production variable as determined by equation prior
        rhs_series = planner._var[product_name][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=product_name,
                            constr_type='==', rhs_var_id_series=rhs_series)

        coeff_val = (planner._intlength / 60) * 10**3 * (1 / (self._sec_system))

        create_constr_lhs_on_interval(planner, constr_name=product_name,
                                        constr_rhs=planner._constr_rhs_dynamic[product_name],
                                        coefficient_map={load_name: coeff_val})


    def _set_h2_production_variable(self):
        """
        Creates a timeseries of Hydrogen Production variables for the optimiser determined by a variable specific
        energy consumption (kWh/kg) relationship. The determined y-variable (Hydrogen Production [kg]) is resolved by
        a Special Ordered Set (SOS) Type 2 in the optimisation solver. In essence, a piecewise linear approximation of
        the defined `_sec_variable_points`. 
        
        The fundamental equation converting SEC [kWh/kg] to [kg] is defined by:
        h2_production [kg] = (1 / sec_system) [kg/kWh] * 10^3 [kWh/MWh] * h2e_load [MW] * \
            (Interval Length [min] / 60) [h]
        """
        planner = self._system_plan
        product_name = self._id + '-h2_produced'
        load_name = self._id + '-mw_load'

        # Calculate sec_variable_points in terms of x: h2e_load [MW] and y: h2_volume [kg]
        self._sec_variable_points['h2e_load'] = self._sec_variable_points['h2e_load_pct'] * self._capacity
        
        self._sec_variable_points['sec'] = self._sec_variable_points['nominal_sec_pct'] * self._sec_system

        self._sec_variable_points['h2_volume'] = self._sec_variable_points['h2e_load'] * (planner._intlength / 60) \
            * 10**3 * (1 / self._sec_variable_points['sec'])
        
        # Extract x, y sample points from user defined _sec_variable_points
        x_samples = self._sec_variable_points['h2e_load'].to_list()
        y_samples = self._sec_variable_points['h2_volume'].to_list()

        # Declare Variables
        ## x_var: h2e_load already declared
        ## y_var:
        create_timeseries_vars(planner, var_name=product_name, lb=0.0, ub=np.inf)

        ## weight vars:
        for idx in range(len(x_samples)):
            create_timeseries_vars(planner, var_name=self._id+'-sec_w'+str(idx), lb=0, ub=1)

        # Define weights and link constraints in solver
        create_sos_type_2(planner, x_samples, y_samples, sos_name=self._id+'-sec', weight_var_name=self._id+'-sec_w',
            x_var_name=load_name, y_var_name=product_name)


    def _price_h2_production(self):
        """
        Defines a price for Hydrogen production [$/kg] as a benefit (not cost) in the optimisation solver.
        """
        planner = self._system_plan
        product_name = self._id + '-h2_produced'
        create_objective_cost(planner, var_name=product_name, decision_var_series=planner._var[product_name], \
            cost=-1*self._h2_price_kg)


    def _set_h2_production_sum(self):
        """Add a variable summating all interval data for the Electrolyser h2 production variables.
        """
        planner = self._system_plan
        name_h2_sum = self._id + '-h2_produced_sum'
        name_h2 = self._id + '-h2_produced'

        # New variable for summation of load series
        create_var(planner, var_name=name_h2_sum, lb=0.0, ub=np.inf)
        
        # New constr with variable as RHS
        rhs_var_id = planner._var[name_h2_sum]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(planner, constr_name=name_h2_sum, constr_type='==', rhs_var_id=rhs_var_id)

        # New lhs summation of constr
        create_constr_lhs_on_interval_fixed_rhs(planner, constr_name=name_h2_sum,
                                                    constr_rhs=planner._constr_rhs_dynamic[name_h2_sum],
                                                    decision_vars=planner._var, coefficient_map={name_h2: 1})


    # =================== Electrolyser MSL & Ramp Functions ====================
    def _set_h2_msl_limit(self, viol_cost=1000.0, Mvalue=1000.0):
        planner = self._system_plan
        par_name = self._id + '-msl_'
        load_name = self._id + '-mw_load'

        # Limit H2 MW load operation to above self._minload
        # Create continuous violation variable and binary indicators
        create_timeseries_vars(planner, var_name=par_name+'violate', lb=0, ub=np.inf)
        create_timeseries_vars(planner, var_name=par_name+'penalise', lb=0, ub=1, solver_type='binary')
        create_timeseries_vars(planner, var_name=par_name+'relieve', lb=0, ub=1, solver_type='binary')

        # Create penalty and relief costs
        create_objective_cost(planner, var_name=par_name+'penalise',
                              decision_var_series=planner._var[par_name+'penalise'],
                              cost=viol_cost)
        create_objective_cost(planner, var_name=par_name+'relieve',
                              decision_var_series=planner._var[par_name+'relieve'],
                              cost=-1*viol_cost)

        # 0 to MSL violation region
        create_constr_rhs_on_interval(planner, constr_name=par_name+'violate', constr_type='>=',
                                      rhs_value=self._minload)
        create_constr_lhs_on_interval(planner, constr_name=par_name+'violate',
                                          constr_rhs=planner._constr_rhs[par_name+'violate'],
                                          coefficient_map={load_name: 1, par_name+'violate': 1})

        # Relax condition for electrolyser off state (P=0)
        create_constr_bigM_on_interval(planner, constr_name=par_name+'penalise',
                                       lhs_vars=planner._var[par_name+'violate'],
                                       rhs_bin_vars=planner._var[par_name+'penalise'],
                                       constr_type='<=', mode='normal', Mvalue=Mvalue)
        create_constr_bigM_on_interval(planner, constr_name=par_name+'relieve',
                                       lhs_vars=planner._var[load_name],
                                       rhs_bin_vars=planner._var[par_name+'relieve'],
                                       constr_type='<=', mode='inverse', Mvalue=Mvalue)


    def _set_ramp_variable(self):
        """Set the ramp variable

        .. note:: The units of the ramp variable are in MW. The user should consider this and dispatch interval length
                    for conventional reporting in MW/min. 
        """
        planner = self._system_plan
        load_name = self._id + '-mw_load'
        ramp_name = self._id + '-mw_ramp'
        down_name = self._id + '-mw_ramp_down'
        up_name = self._id + '-mw_ramp_up'

        timeless1 = pd.DataFrame(index=planner._timeseries[:-1], data={'interval':range(planner._n-1)})
        create_period_vars(planner, ramp_name, timeless1, lb=0, ub=np.inf)

        # Set the variable equal to the abs value difference between intervals 
        rhs_series = planner._var[ramp_name][['interval', 'variable_id']]

        # Ramp down
        create_constr_rhs_on_interval_dynamicvar(planner, down_name, constr_type="<=", \
            rhs_var_id_series=rhs_series, intervals=timeless1['interval'].to_list())
        offset_matrix = pd.DataFrame(data={load_name: [1,-1]},index=[0,1])
        create_contr_lhs_on_interval_matrix(planner, constr_name=down_name, constr_rhs=planner._constr_rhs_dynamic[down_name], \
            coeff_matrix=offset_matrix)

        # Ramp up
        create_constr_rhs_on_interval_dynamicvar(planner, up_name, constr_type="<=", \
            rhs_var_id_series=rhs_series, intervals=timeless1['interval'].to_list())
        offset_matrix = pd.DataFrame(data={load_name: [-1,1]},index=[0,1])
        create_contr_lhs_on_interval_matrix(planner, constr_name=up_name, constr_rhs=planner._constr_rhs_dynamic[up_name], \
            coeff_matrix=offset_matrix)


    def _set_ramp_cost(self, cost=10.0):
        planner = self._system_plan
        ramp_name = self._id+'-mw_ramp'

        create_objective_cost(planner, ramp_name, decision_var_series=planner._var[ramp_name], cost=cost)


    # ================ Electrolyser Production Target Functions ================
    def _set_force_h2_load(self, mw_value):
        planner = self._system_plan
        const_name = self._id + '-mw_force'
        load_name = self._id + '-mw_load'
        create_constr_rhs_on_interval(planner, constr_name=const_name, constr_type='==', rhs_value=mw_value)
        create_constr_lhs_on_interval(planner, constr_name=const_name, constr_rhs=planner._constr_rhs[const_name],
                                      coefficient_map={load_name: 1}) 


    def _set_production_target(self, target_value=100, bound="max", period="hour"):
        """Set production target

        .. warning:: In development (from old version)
        """
        planner = self._system_plan

        # Create variable production target, set equal to value
        # Set sum of lhs params equal to, greater than or less than variable on rhs dynamic

        # consider what happens if timseries data input is hourly but requests < hourly period
        # consider what happens if timeseries data stops midway through week, how to consider last week?
        
        # Create a production target variable on selected intervals determined by `period`

        # production target == p1 + p2 + p3
        original = pd.DataFrame(index=planner._timeseries, data={'interval':range(planner._n)})
        original['TimeBegin'] = original.index - timedelta(minutes=planner._intlength)

        if period == "interval":
            newseries = original.copy()
            create_timeseries_vars(planner, f'production_target_{bound}_{period}', lb=0, ub=np.inf)
        elif period == "hour":
            newseries = original.copy().resample('H',closed='right',label='right').max()
            create_period_vars(planner, f'production_target_{bound}_{period}', newseries, lb=0, ub=np.inf)
        elif period == "day":
            newseries = original.copy().resample('D',closed='right').max()
            create_period_vars(planner, f'production_target_{bound}_{period}', newseries, lb=0, ub=np.inf)
        elif period == "week":
            newseries = original.set_index('TimeBegin').resample('W',closed='right').max()
            create_period_vars(planner, f'production_target_{bound}_{period}', newseries, lb=0, ub=np.inf)
        elif period == "month":
            newseries = original.set_index('TimeBegin').resample('M',closed='right').max()
            create_period_vars(planner, f'production_target_{bound}_{period}', newseries, lb=0, ub=np.inf)
        elif period == "year":
            newseries = original.set_index('TimeBegin').resample('A',closed='right').max()
            create_period_vars(planner, f'production_target_{bound}_{period}', newseries, lb=0, ub=np.inf)
        else:
            raise Exception("Invalid period passed. Valid periods: ['interval','hour','day','week','month','year']")

        # Check Bound Equality
        if bound == "max":
            equality = "<="
        elif bound == "min":
            equality = ">="
        elif bound == "equal":
            equality = "=="
        else:
            raise Exception("bound variable must be either: max, min, equal")

        # Set the production target variables to user defined value
        # e.g. production target >= 100
        # TODO: MAKE ELASTIC
        create_constr_rhs_on_interval(planner, f'prd_tgt_{bound}_{period}', constr_type=equality, \
            rhs_value=target_value, intervals=newseries['interval'].to_list())
        
        create_constr_lhs_on_interval(planner, f'prd_tgt_{bound}_{period}', \
            constr_rhs=planner._constr_rhs[f'prd_tgt_{bound}_{period}'], \
            coefficient_map={f'production_target_{bound}_{period}': 1})


        # Set the production target variables to be the summation of DI production outputs.
        rhs_series = planner._var[f'production_target_{bound}_{period}'][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, f'sum_prd_tgt_{bound}_{period}', constr_type="==", \
            rhs_var_id_series=rhs_series, intervals=newseries['interval'].to_list())

        create_constr_lhs_sum_up_to_intervals(planner, constr_name=f'sum_prd_tgt_{bound}_{period}', \
            constr_rhs=planner._constr_rhs_dynamic[f'sum_prd_tgt_{bound}_{period}'], \
            coefficient_map={self._id+'-h2_produced': 1})


        # Configuration is..
        # var: production_target_{bound}_{period} >= defined_value
        #  sum(h2_production) = RHS production_target...

        # To make this elastic add param in prd_tgt equality constraint

      
    # ======================= Hydrogen Storage Functions =======================
    def _set_h2_production_tracking(self):
        planner = self._system_plan
        stor_name = self._id + '-h2_stored'
        prod_name = self._id+'-h2_produced'

        # Var to reflect tracking sum of h2 produced.
        create_timeseries_vars(planner, var_name=stor_name, lb=0, ub=np.inf)

        # Constr: h2_stored = h2_stored[t-1] + h2_production[t]
        rhs_series = planner._var[stor_name][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=stor_name, constr_type='==', \
            rhs_var_id_series=rhs_series)

        # Create LHS on interval-shifting matrix
        coeff_matrix = pd.DataFrame(data={stor_name: [1,np.nan], prod_name: [np.nan,1]}, index=[-1,0])
        create_contr_lhs_on_interval_matrix(planner, constr_name=stor_name, \
            constr_rhs=planner._constr_rhs_dynamic[stor_name], coeff_matrix=coeff_matrix)


    def _set_h2_storage_initial(self):
        planner = self._system_plan
        stor_name = self._id + '-h2_stored'
        init_name = self._id + '-h2_stored_initial'

        first_interval = pd.DataFrame(data={'interval':[0]})
        create_period_vars(planner, var_name=init_name, interval_set=first_interval, lb=0, ub=np.inf)
        create_constr_rhs_on_interval(planner, constr_name=init_name, constr_type='==', \
            rhs_value=self._storage_initial, intervals=[0])
        create_constr_lhs_on_interval(planner, constr_name=init_name, \
            constr_rhs=planner._constr_rhs[init_name], coefficient_map={init_name: 1})

        # Add initial value as parameter to LHS of storage amount
        create_constr_lhs_on_interval(planner, constr_name=stor_name, \
            constr_rhs=planner._constr_rhs_dynamic[stor_name], coefficient_map={init_name: 1})


    def _set_h2_storage_final(self):
        planner = self._system_plan
        stor_name = self._id + '-h2_stored'
        final_name = self._id + '-h2_stored_final'
        force_final = self._id + '-h2_stored_final_constraint'

        final_interval = pd.DataFrame(data={'interval':[planner._n - 1]})
        create_period_vars(planner, var_name=final_name, interval_set=final_interval, lb=0, ub=np.inf)
        create_constr_rhs_on_interval(planner, constr_name=final_name, constr_type='==', \
            rhs_value=self._storage_final, intervals=[planner._n - 1])
        create_constr_lhs_on_interval(planner, constr_name=final_name, \
            constr_rhs=planner._constr_rhs[final_name], coefficient_map={final_name: 1})

        # New constraint to force 'h2_stored_final' to equal 'h2_storage_amount' in last interval
        rhs_series = planner._var[stor_name][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=force_final, constr_type="==", \
            rhs_var_id_series=rhs_series, intervals=[planner._n - 1])

        create_constr_lhs_on_interval(planner, constr_name=force_final, \
            constr_rhs=planner._constr_rhs_dynamic[force_final], coefficient_map={final_name: 1})


    def _set_h2_storage_max(self):
        planner = self._system_plan
        stor_name = self._id + '-h2_stored'
        max_name = self._id + '-h2_stored_limit'
        
        create_constr_rhs_on_interval(planner, constr_name=max_name, constr_type='<=', \
            rhs_value=self._storage_max)

        create_constr_lhs_on_interval(planner, constr_name=max_name, \
            constr_rhs=planner._constr_rhs[max_name], coefficient_map={stor_name: 1})


    def _set_storage_external_flow(self, external_flow=-500):
        """Set storage external flow

        .. todo:: Add dynamic storage external_flow that varies between intervals
        
        """
        planner = self._system_plan
        stor_name = self._id + '-h2_stored'
        ext_name = self._id + '-h2_stored_extflow'
        self._storage_ext_flow = [external_flow] * planner._n

        # Create external in/out-flow variable and set to defined mass-flow value.
        create_timeseries_vars(planner, var_name=ext_name, lb=-np.inf, ub=np.inf)
        create_constr_rhs_on_interval(planner, constr_name=ext_name, constr_type="==", \
            rhs_value=self._storage_ext_flow)
        create_constr_lhs_on_interval(planner, constr_name=ext_name, \
            constr_rhs=planner._constr_rhs[ext_name], coefficient_map={ext_name: 1})

        # Add external in/out-flow variable as LHS argument to storage system
        create_constr_lhs_on_interval(planner, constr_name=stor_name, \
            constr_rhs=planner._constr_rhs_dynamic[stor_name], coefficient_map={ext_name: 1})




# Backend API imports
from nemglo.lite import *
from nemglo.defaults import *
# import validate as inv
import argparse
import sys

# Flask/CORS imports
from flask import Flask, request, jsonify, abort
from flask_cors import CORS, cross_origin
# app = Flask(__name__)

# Flask
if getattr(sys, 'frozen', False):
    template_folder = os.path.join(sys._MEIPASS, 'templates')
    static_folder = os.path.join(sys._MEIPASS, 'static')
    app = Flask(__name__, template_folder=template_folder, static_folder=static_folder)
else:
    app = Flask(__name__)

# Flask cors
cors = CORS(app)
app.config['CORS_HEADERS'] = 'Content-Type'

# Return error log for API request
from flask import abort as fabort, make_response
def abort(status_code, message):
    response = make_response(f'{status_code}\n{message}')
    response.status_code = status_code
    fabort(response)


def throw_error_log():
    with open(LOG_FILEPATH, 'r') as file:
        logfile = file.read()
    err_message = logfile.split('200 -')[-1:]
    abort(500, err_message[0])


# API - Tests [preview]
@app.route('/', methods=['GET'])
def default():
    return "you have reached nemglo"


@app.route('/api/', methods=['GET'])
def default_api():
    return "NEMGLO API"


# API - Market Data inputs [preview]
@app.route('/api/get-market-data', methods=['POST'])
def api_get_market_data():
    try:
        conf = request.json
        return get_market_data(conf)
    except:
        # with open(LOG_FILEPATH, 'r') as file:
        #     logfile = file.read()
        # err_message = logfile.split('200 -')[-1:]
        # abort(500, err_message[0])
        throw_error_log()


# API - Generator Data inputs [preview]. ONLY FOR SINGLE GEN, CANNOT PASS BOTH
@app.route('/api/get-generator-data', methods=['POST'])
def api_get_generator_data():
    try:
        conf = request.json
        return get_generator_data(conf)
    except:
        # with open(LOG_FILEPATH, 'r') as file:
        #     logfile = file.read()
        # err_message = logfile.split('200 -')[-1:]
        # abort(500, err_message[0])
        throw_error_log()


# API - Model Simulation [MAIN]
@app.route('/api/get-data', methods=['POST'])
def api_get_data():
    try:
        conf = request.json
        print(conf)
        return get_operation(conf)
    except:
        # with open(LOG_FILEPATH, 'r') as file:
        #     logfile = file.read()
        # err_message = logfile.split('200 -')[-1:]
        # abort(500, err_message[0])
        throw_error_log()


def run(port=5000):
    print("\n======================================================================\n"+\
          "Access NEMGLO-app (web interface) at: https://www.nemglo.org/simulator \n"+\
          "======================================================================\n")
    app.run(port=port)

if __name__=='__main__':
        
    # Check if cache folder provided and valid filepath
    parser = argparse.ArgumentParser()
    parser.add_argument('--cache', '-c', type=str, \
        help="provide a local filepath to a folder to be used for caching data")
    args = parser.parse_args()

    # Determine cache folder
    if (args.cache is None):
        logging.info("Default data cache location is: {}.".format(DATA_CACHE.FILEPATH))
    elif (not os.path.exists(args.cache)):
        logging.info("Default data cache location is: {}.".format(DATA_CACHE.FILEPATH))
    else:
        DATA_CACHE.update_path(args.cache)
        logging.info("Updated preffered data cache location to: {}."+ \
            "Note, log files will save to default cache.".format(DATA_CACHE.FILEPATH))

    run()


import logging
from .defaults import *
import os

# Make default cache location for logging, if not existing
if not os.path.exists("CACHE"):
    os.mkdir("CACHE")

# Log File Config
logging.basicConfig(
     filename=LOG_FILEPATH,
     filemode="w+",
     level=logging.INFO, 
     format= '[%(asctime)s] {%(filename)s:%(lineno)d} [%(levelname)s]: %(message)s',
     datefmt='%H:%M:%S',
 )

# Log Console Config
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
console.setFormatter(formatter)
logging.getLogger('').addHandler(console)

# Disable other package loggers (min Warning level)
logging.getLogger("mip").setLevel(logging.WARNING)

# Commence Logging
logger = logging.getLogger(__name__)
logger.info(f"Commence Logging. Saved to file: {LOG_FILEPATH}")

# Must define logging config before remaining package imports
from nemglo.lite import *
from nemglo.api import * 
from nemglo.planning import data_fetch, planner
from nemglo.components import electrolyser, emissions, renewables
from nemglo.defaults import *

from importlib.metadata import version
try:
    __version__ = version("nemglo")
    logging.info("Running NEMGLO version: {}.".format(__version__))
except:
    pass





# Execution variables
import os
class CACHE():
    def __init__(self, FILEPATH):
        self.FILEPATH = FILEPATH

    def update_path(self, FILEPATH):
        self.FILEPATH = FILEPATH

DATA_CACHE = CACHE(os.path.join(os.getcwd(), "CACHE"))
LOG_FILEPATH = os.path.join(DATA_CACHE.FILEPATH, "nemglo_sim.log")

# Naming conventions
variable_names = ['vre_cap']
constraint_names = []

cost_variables_shadow = ['impact_emissions', 'relief_prd_tgt', 'msl_penalise', 'msl_relieve']
cost_variables_realized = ['h2e_load', 'h2_production', 'ppa_vre']

# Electrolyser Default SEC variable profiles and characteristics
# Beerbühl et.al, 2015, https://doi.org/10.1016/j.ejor.2014.08.039
SEC_PROFILE_AE = {'h2e_load_pct': [0, 0.2, 0.4, 0.6, 0.8, 1.0],
                    'nominal_sec_pct': [0.8068, 0.8068, 0.8181, 0.8523, 0.9091, 1.0]}
# Kopp et.al, 2017, https://doi.org/10.1016/j.ijhydene.2016.12.145
SEC_PROFILE_PEM = {'h2e_load_pct': [0, 0.2, 0.4, 0.6, 0.8, 1.0],
                    'nominal_sec_pct': [0.8383, 0.8383, 0.8598, 0.8982, 0.9315, 1.0101]}
SEC_NOMINAL_AE = 50
SEC_NOMINAL_PEM = 66
SEC_CONVERSION_AE = 1.0
SEC_CONVERSION_PEM = 1.0

PALETTE = ['#b2d4ee','#849db1','#4f6980',
           '#B4E3BC','#89AE8F','#638b66',
           '#ffb04f','#de9945','#af7635',
           '#ff7371','#d6635f','#b65551',
           '#AD134C','#cc688d','#ff82b0']


import pandas as pd
import numpy as np
from nemglo.planning.planner import Plan
from nemglo.backend import input_validation as inv
from nemglo.backend.optimiser_formatters import *

class Generator:
    """Generator class is used to create an object to store parameters defining an Variable Renewable Energy (VRE)
    source for the load to contract a Power Purchase Agreements (PPA) with.

    Parameters
    ----------
    system_plan : nemglo.planner.Plan
        The system plan object for which the Generator should be linked to.
    
    identifier : str
        A unique identifier to refer to the object as. For example, 'VRE1' for the first PPA, 'VRE2' for the second.
    
    .. todo:: Inherit data directly via data_fetch and flag an attribute to signify this vs custom inputs.
        From using python, writing a yaml file to show inputs saved in simulation.

    """
    def __init__(self, system_plan, identifier):
        # Error check on parsed arguments
        assert isinstance(system_plan, Plan), "Generator Argument: 'system_plan' must be nemglo.planner.Plan object"
        assert isinstance(identifier, str), "Generator Argument: 'identifier' must be a str"
        inv.validate_unique_id(self.__class__, system_plan, identifier)

        # Link object to Plan
        self._system_plan = system_plan
        cname = self.__class__.__name__
        if cname in self._system_plan._components:
            self._system_plan._components.update({cname: self._system_plan._components[cname] + [self]})
        else:
            self._system_plan._components.update({cname: [self]})
        self._id = identifier

        # Generic Plant Characteristics
        self._duid = None
        self._capacity = None
        self._trace = None

        # PPA Structure
        self._ppa_strike = None
        self._ppa_floor = None

    # Functions to validate inputs
    def load_vre_parameters(self, duid, capacity, trace, ppa_strike, ppa_floor=None):
        """Load VRE parameters defining the Generator object.

        Parameters
        ----------
        duid : str
            The unique string of the VRE Generator. If using a custom input, create a custom name for reference.
        capacity : float
            The desired rated capacity of the VRE Generator, in MW. Note the input trace will be scaled accordingly.
        trace : pandas.DataFrame
            Timeseries data of generator trace.

            ========  ================================================
            Columns:  Description:
            Time      Dispatch interval timestamps (as `datetime.datetime`)
            {duid}	  Capacity factor trace values between 0 and 1, % (as `np.float64`)
            ========  ================================================
        
        ppa_strike : float
            The strike price for the Power Purchase Agreement (PPA) with the VRE Generator, in $/MWh.
        ppa_floor : float, optional
            A floor price for the PPA contract if desired, in $/MWh, by default None.
        
        """
        # Validate Inputs
        assert isinstance(duid, str), "Generator Argument: 'duid' must be a str"
        assert isinstance(capacity, float), "Generator Argument: 'capacity' must be a float"
        assert isinstance(trace, pd.DataFrame), "Generator Argument: 'trace' must be a pd.DataFrame"
        assert isinstance(ppa_strike, float), "Generator Argument: 'ppa_strike' must be a float"
        assert isinstance(ppa_floor, (float, type(None))), "Generator Argument: 'ppa_floor' must be a \
            (float, or None)"

        self._validate_trace(trace, duid)
    
        # Store Values
        self._duid = duid
        self._capacity = capacity
        self._trace = trace
        self._ppa_strike = ppa_strike
        self._ppa_floor = ppa_floor


    def _validate_trace(self, trace, duid):
        schema = inv.DataFrameSchema(name="trace", primary_keys=['Time', duid])
        schema.add_column(inv.SeriesSchema(name='Time', data_type=np.dtype('datetime64[ns]'), no_duplicates=True, \
            ascending_order=True, minimum=self._system_plan._timeseries[0], maximum=self._system_plan._timeseries[-1]))
        schema.add_column(inv.SeriesSchema(name=duid, data_type=np.float64, must_be_real_number=True, \
            not_negative=True))
        schema.validate(trace)


    # Functions to call from within the planner
    def add_ppa_contract(self):
        """Constructs a PPA contract between the load and VRE Generator, given the defined parameters of the generator.

        .. tip:: This function is for standard python use. For more advanced control / direct interfacing, instead use
            the `_set` functions.
        """
        # Verify input parameters exist
        assert (self._duid and self._capacity and any(self._trace) and self._ppa_strike), \
            "Missing Generator Inputs. Be sure to call `load_parameters` before `add_`"

        # Define capacity variable
        inv.validate_existing(self._system_plan, self._id+'-vre_cap', as_var=True, as_constr_lhs=True, \
            as_constr_rhs=True)
        self._set_capacity_var()
        self._set_fixed_capacity()

        # Define VRE availability
        inv.validate_existing(self._system_plan, self._id+'-vre_avail', as_var=True, as_constr_lhs=True, \
            as_constr_rhs_dynamic=True)
        self._set_availability()

        # Define PPA cost function
        inv.validate_existing(self._system_plan, self._id+'-ppa_cfd', as_objective_cost=True)
        self._set_ppa_cost_function()

        # Define PPA volume sum for RECs
        inv.validate_existing(self._system_plan, self._id+'-ppa_rec_sum', as_var=True, as_constr_lhs=True, \
            as_constr_rhs_dynamic=True)
        self._set_ppa_rec_sum()


    def remove_ppa_contract(self):
        """Removes all assosciations of the Generator object from the Plan object.
        """
        remove_element(self._system_plan, self._id, all_assoc=True)


    def _set_capacity_var(self, ub=np.inf):
        """Defines Generator capacity variable in the optimiser as -> Capacity = 100.0
        """
        name_cap = self._id + '-vre_cap'
        planner = self._system_plan

        create_var(planner, var_name=name_cap, lb=0, ub=ub)


    def _set_fixed_capacity(self):
        name_cap = self._id + '-vre_cap'
        planner = self._system_plan

        create_constr_rhs(planner, constr_name=name_cap, constr_type='==', rhs_value=self._capacity)
        
        create_constr_lhs_on_interval(planner, constr_name=name_cap, \
            constr_rhs=planner._constr_rhs[name_cap], coefficient_map={name_cap: 1})
        

    def _set_availability(self):
        """Defines the VRE generator's availability in the optimiser.
        """
        name_avail = self._id + '-vre_avail'
        name_cap = self._id + '-vre_cap'
        planner = self._system_plan

        create_timeseries_vars(planner, var_name=name_avail, lb=0.0, ub=np.inf)
        
        var_ids = planner._var[name_avail][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(planner, constr_name=name_avail,
            constr_type='==', rhs_var_id_series=var_ids)

        create_constr_lhs_on_interval_dynamic(planner, constr_name=name_avail,
            constr_rhs=planner._constr_rhs_dynamic[name_avail],
            decision_vars=planner._var[name_cap],
            coefficient_list=self._trace[self._duid])


    def _set_ppa_cost_function(self):
        """Add the PPA cost function of the VRE generator as a CfD traded on the variable volume of available
        generation.

        .. note:: The cost component in the objective function is expressed in $/MW.
        """
        planner = self._system_plan
        strike = self._ppa_strike
        spotprices = planner._prices
        intlength = planner._intlength
        n = planner._n
        floor = self._ppa_floor
        name_avail = self._id + '-vre_avail'

        if floor != None:
            # PPA takes the difference between strike and floor if spot < floor.
             cfd_val = [(strike - max(floor, spotprices[i])) * (intlength/60) for i in range(0, len(spotprices))]
        else:
            # PPA takes the difference between strike and spot.
            cfd_val = [(strike - spotprices[i]) * (intlength/60) for i in range(0, len(spotprices))]

        series = planner._var[name_avail]
        price_series = pd.DataFrame({'interval': range(n), 'cost': cfd_val})

        create_objective_cost(planner, var_name=self._id+'-ppa_cfd', decision_var_series=series, cost=price_series)


    def _set_ppa_rec_sum(self):
        """Add a summation of the traded volume of the PPA CfD which can used to consider the volume of Renewable Energy
        Certificates received by the load on the assumption of a bundled PPA.

        .. note:: The `ppa_rec_sum` variable is expressed in MW.
        """
        planner = self._system_plan
        name_rec_sum = self._id + '-ppa_rec_sum'
        name_avail = self._id + '-vre_avail'

        # New variable for summation of vre_avail series
        create_var(planner, var_name=name_rec_sum, lb=0.0, ub=np.inf)

        # New constr with variable as RHS
        rhs_var_id = planner._var[name_rec_sum]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(planner, constr_name=name_rec_sum, constr_type='==', rhs_var_id=rhs_var_id)

        # New lhs summation of constraint
        create_constr_lhs_on_interval_fixed_rhs(planner, constr_name=name_rec_sum,
                                                constr_rhs=planner._constr_rhs_dynamic[name_rec_sum],
                                                decision_vars=planner._var,
                                                coefficient_map={name_avail: 1})



import pandas as pd
import numpy as np
import os
import json
from datetime import datetime
from nemglo.backend.optimiser_formatters import *
from nemglo.backend.solver_operator import Opt_Solver as solver
from nemglo.planning.data_fetch import *
from nemglo.defaults import *

from nemglo.backend import input_validation as inv

class Plan():
    """ Main Planner object to create and run the load optimisation.

    """
    def __init__(self, identifier):

        # Error check on parsed arguments
        assert isinstance(identifier, str), "Plan Argument: 'identifier' must be a str"
        
        # Object Parameters
        self._id = identifier
        self._components = {}

        # Market Data
        self._prices = []
        self._lgc_price = None
        self._timeseries = []
        self._n = None # KEEP AS IS
        self._intlength = None # KEEP AS IS
        self._rec_price = None

        # # Market => VRE Data
        # self._vre_det = None
        # self._vre_cf_1 = []
        # self._vre_cf_2 = []

        # Optimiser Data
        self._solver = 'GRB'
        self._var = {}
        self._objective_cost = {}
        self._constr_lhs = {}
        self._constr_rhs = {}
        self._constr_rhs_dynamic = {}
        self._constr_bigM = {}
        self._sos_2 = {}
        self._varid_idx = 0
        self._conid_idx = 0

        # Result Data
        self._out_vars = None
        self._out_constr = None



    def load_market_prices(self, market_prices):
        """
        Load market inputs to planner module.

        Parameters
        ----------
        timeseries : bool
                List of timestamps defining the intervals in the simulation period. The default is True.

        market_prices : pandas.DataFrame
            Timeseries data of energy prices to consider in optimisation.

            ========  ================================================
            Columns:  Description:
            Time      Dispatch interval timestamps (as `datetime.datetime`)
            Prices	  Market energy prices, $/MWh (as `np.float64`)
            ========  ================================================
        
        """

        self._validate_market_prices(market_prices)

        if any([self._timeseries, self._prices]):
            print("WARNING: Overwriting existing market_prices")

        self._timeseries = market_prices['Time'].to_list() #timeseries
        self._prices = market_prices['Prices'].to_list() #prices
        self._n = len(self._timeseries)
        self._intlength = np.diff(self._timeseries).min().seconds / 60


    def _validate_market_prices(self, market_prices):
        schema = inv.DataFrameSchema(name="market_prices", primary_keys=['Time', 'Prices'])
        schema.add_column(inv.SeriesSchema(name='Time', data_type=np.dtype('datetime64[ns]'), no_duplicates=True, ascending_order=True))
        schema.add_column(inv.SeriesSchema(name='Prices', data_type=np.float64, must_be_real_number=True))
        schema.validate(market_prices)


    def _load_rec_price(self, rec_price):
        assert isinstance(rec_price, (float, pd.DataFrame)), "load_rec_price Argument: 'rec_price' must be a float or pd.DataFrame"

        if type(rec_price) == pd.DataFrame:
            self._validate_certificate_prices(rec_price)

        self._rec_price = rec_price


    def _validate_certificate_prices(self, certificate_prices):
        schema = inv.DataFrameSchema(name="certificate_prices", primary_keys=['Time', 'Prices'])
        schema.add_column(inv.SeriesSchema(name='Time', data_type=np.dtype('datetime64[ns]'), no_duplicates=True, \
            ascending_order=True, minimum=self._timeseries[0], maximum=self._timeseries[-1]))
        schema.add_column(inv.SeriesSchema(name='REC_price', data_type=np.float64, must_be_real_number=True))
        schema.validate(certificate_prices)


    def _price_mkt_rec(self, allow_buying=True, allow_selling=False):
        name_rec = self._id + '-mkt_rec_buy'
        name_rec_sell = self._id + '-mkt_rec_sell'
       
        # Price series depending on float or df
        cost = self._rec_price
        if type(cost) == pd.DataFrame:
            cost['interval'] = range(self._n)
            cost.rename(columns={'REC_price': 'cost'}, inplace=True)
            cost['cost'] = cost['cost'].mul(self._intlength/60)
        else:
            cost = cost * (self._intlength/60)

        if allow_buying:
            create_timeseries_vars(self, var_name=name_rec, lb=0.0, ub=np.inf)
            create_objective_cost(self, var_name=name_rec, decision_var_series=self._var[name_rec],
                                    cost=cost)

        if allow_selling:
            create_timeseries_vars(self, var_name=name_rec_sell, lb=0.0, ub=np.inf)
            create_objective_cost(self, var_name=name_rec_sell, decision_var_series=self._var[name_rec_sell],
                                    cost=(-1 * cost))

    
    def _set_mkt_rec_sum(self, allow_buying=True, allow_selling=False):
        name_rec = self._id + '-mkt_rec_buy'
        name_rec_sum = self._id + '-mkt_rec_buy_sum'
        name_rec_sell = self._id + '-mkt_rec_sell'
        name_rec_sell_sum = self._id + '-mkt_rec_sell_sum'

        if allow_buying:
            # New variable for summation of vre_avail series
            create_var(self, var_name=name_rec_sum, lb=0.0, ub=np.inf)

            # New constr with variable as RHS
            rhs_var_id = self._var[name_rec_sum]['variable_id'].values[0]
            create_constr_rhs_dynamicvar(self, constr_name=name_rec_sum, constr_type='==', rhs_var_id=rhs_var_id)

            # New lhs summation of constraint
            create_constr_lhs_on_interval_fixed_rhs(self, constr_name=name_rec_sum,
                                                    constr_rhs=self._constr_rhs_dynamic[name_rec_sum],
                                                    decision_vars=self._var,
                                                    coefficient_map={name_rec: 1})

        if allow_selling:
            # New variable for summation of vre_avail series
            create_var(self, var_name=name_rec_sell_sum, lb=0.0, ub=np.inf)

            # New constr with variable as RHS
            rhs_var_id = self._var[name_rec_sell_sum]['variable_id'].values[0]
            create_constr_rhs_dynamicvar(self, constr_name=name_rec_sell_sum, constr_type='==', rhs_var_id=rhs_var_id)

            # New lhs summation of constraint
            create_constr_lhs_on_interval_fixed_rhs(self, constr_name=name_rec_sell_sum,
                                                    constr_rhs=self._constr_rhs_dynamic[name_rec_sell_sum],
                                                    decision_vars=self._var,
                                                    coefficient_map={name_rec_sell: 1})


    def _price_mkt_rec_sum(self, allow_buying=True, allow_selling=False):
        name_rec_sum = self._id + '-mkt_rec_buy_sum'
        name_rec_sell_sum = self._id + '-mkt_rec_sell_sum'

        cost = self._rec_price * (self._intlength/60)

        if allow_buying:
            create_var(self, var_name=name_rec_sum, lb=0.0, ub=np.inf)
            create_objective_cost(self, var_name=name_rec_sum, decision_var_series=self._var[name_rec_sum],
                                    cost=cost)

        if allow_selling:
            create_var(self, var_name=name_rec_sell_sum, lb=0.0, ub=np.inf)
            create_objective_cost(self, var_name=name_rec_sell_sum, decision_var_series=self._var[name_rec_sell_sum],
                                    cost=(-1 * cost))


    def _set_acquire_rec_on_interval(self, allow_buying=True, allow_selling=False):
        name_rec = self._id + '-mkt_rec_buy'
        name_rec_sell = self._id + '-mkt_rec_sell'
        name_load = self._components['Electrolyser'][0]._id + '-mw_load'
        name_constr = self._id+'-rec_acquire'

        coeffs = {}
        if allow_buying:
            coeffs.update({name_rec: 1})

        if allow_selling:
            coeffs.update({name_rec_sell: -1})

        if 'Generator' in self._components:
            for gen in self._components['Generator']:
                name_ppa_rec = gen._id + '-vre_avail'
                coeffs.update({name_ppa_rec: 1})

        rhs_series = self._var[name_load][['interval', 'variable_id']]
        create_constr_rhs_on_interval_dynamicvar(self, constr_name=name_constr,
                                    constr_type='>=', rhs_var_id_series=rhs_series)

        create_constr_lhs_on_interval(self, constr_name=name_constr, constr_rhs=self._constr_rhs_dynamic[name_constr],
                                        coefficient_map=coeffs)


    def autoadd_rec_price_on_interval(self, rec_price, allow_buying=True, allow_selling=True):
        assert isinstance(allow_buying, bool), "`autoadd_rec_price_on_interval` Argument: 'allow_buying' must be a bool"
        assert isinstance(allow_selling, bool), "`autoadd_rec_price_on_interval` Argument: 'allow_buying' must be a bool"

        self._load_rec_price(rec_price)
        self._price_mkt_rec(allow_buying, allow_selling)
        self._set_mkt_rec_sum(allow_buying, allow_selling)
        self._set_acquire_rec_on_interval(allow_buying, allow_selling)


    def _set_acquire_rec_on_sum(self, allow_buying=True, allow_selling=False):
        name_rec_sum = self._id + '-mkt_rec_buy_sum'
        name_rec_sell_sum = self._id + '-mkt_rec_sell_sum'
        name_load = self._components['Electrolyser'][0]._id + '-mw_load_sum'
        name_constr = self._id+'-rec_acquire'

        coeffs = {}
        if allow_buying:
            coeffs.update({name_rec_sum: 1})

        if allow_selling:
            coeffs.update({name_rec_sell_sum: -1})

        if 'Generator' in self._components:
            for gen in self._components['Generator']:
                name_ppa_rec = gen._id + '-ppa_rec_sum'
                coeffs.update({name_ppa_rec: 1})

        rhs_series = self._var[name_load]['variable_id'].values[0]
        create_constr_rhs_dynamicvar(self, constr_name=name_constr,
                                    constr_type='>=', rhs_var_id=rhs_series)

        create_constr_lhs_on_interval(self, constr_name=name_constr, constr_rhs=self._constr_rhs_dynamic[name_constr],
                                        coefficient_map=coeffs)


    def autoadd_rec_price_on_total(self, rec_price, allow_buying=True, allow_selling=True):
        assert isinstance(allow_buying, bool), "`autoadd_rec_price_on_total` Argument: 'allow_buying' must be a bool"
        assert isinstance(allow_selling, bool), "`autoadd_rec_price_on_total` Argument: 'allow_buying' must be a bool"

        self._load_rec_price(rec_price)
        assert isinstance(self._rec_price, float), "`autoadd_rec_price_on_total` requires a static REC price input as a float type. \
            Instead use `_on_interval` if you wish to define a dynamic REC price"
        self._price_mkt_rec_sum(allow_buying, allow_selling)
        self._set_acquire_rec_on_sum(allow_buying, allow_selling)
       

    def view_variable(self, name):
        """View optimiser variables which are in the planner object.

        Parameters
        ----------
        name : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        view_items = dict((k, self._var[k]) for k in [var for var in self._var if var.__contains__(name)])
        if len(view_items) > 0:
            view_items = pd.concat(view_items)
            view_items = view_items.droplevel(1, axis=0)
            view_items.index.name = 'variable_name'
            view_items.reset_index(inplace=True)
        else:
            print("Variable name not found in Plan: '{}'".format(self._id))
        return view_items
    
    # def view_constraint(self, constraint_name):
    # 	"""_summary_

    # 	Parameters
    # 	----------
    # 	constraint_name : _type_
    # 		_description_

    # 	Returns
    # 	-------
    # 	[name] : pandas.DataFrame
    # 		Timeseries data of generator trace.

    # 		========  ================================================
    # 		Columns:  Description:
    # 		Time      Dispatch interval timestamps (as `datetime.datetime`)
    # 		{duid}	  Capacity factor trace values between 0 and 1, % (as `np.float64`)
    # 		========  ================================================
        
    # 	"""

    # 	# if the interval is not given use (zero or none) 

    # 	# For the normal rhs one:
    # 	# columns: constraint_name, constraint_id, interval, coeff_var_1, var_1_name[var_id_1], .... , type, rhs



    # 	return var

    def add_emissions(self, emissions_obj):

        self._emissions += [emissions_obj]


    def relax_and_price_constr_violation(self, constr_name, sense, cost):
        """
        Need to consider the sense + or - of variable in LHS as pos or neg cost
        """
        # Not advisable to use relief var on equal bound production limit constraints
        if constr_name.__contains__('equal'):
            print("WARNING: It is not advisable to relief and price violation for `equal` bound constraints.\
                The cost implementation is not valid.")

        # Get constr_rhs
        if constr_name in self._constr_rhs_dynamic:
            print("in dynamic")
            sel_df = self._constr_rhs_dynamic[constr_name]
        elif constr_name in self._constr_rhs:
            sel_df = self._constr_rhs[constr_name]
            print("in normal RHS")
        else:
            raise Exception("`price_constraint_violation` could not find constr_name in planner object.\
                Only RHS and RHS_dynamic objects are considered")

        # Create violation variables
        if sense == '+':
            create_period_vars(self, var_name=f"relief_{constr_name}", interval_set=sel_df, lb=0, ub=np.inf)
        elif sense == '-':
            create_period_vars(self, var_name=f"relief_{constr_name}", interval_set=sel_df, lb=-np.inf, ub=0)
        else:
            raise Exception('Invalid `sense` argument. Must be one of '-' or '+'')

        # Insert relief vars into the LHS of constraint
        if constr_name in self._constr_rhs:
            create_constr_lhs_on_interval(self, constr_name=constr_name, constr_rhs=self._constr_rhs[constr_name],\
                coefficient_map={f"relief_{constr_name}": 1})

        if constr_name in self._constr_rhs_dynamic:
            create_constr_lhs_on_interval_dynamic(self, constr_name=constr_name, \
                constr_rhs=self._constr_rhs_dynamic[constr_name], coefficient_map={f"relief_{constr_name}": 1})

        # Price relief vars in objective cost function
        create_objective_cost(self, var_name=f"relief_{constr_name}", \
            decision_var_series=self._var[f"relief_{constr_name}"], cost=cost)

        return


    def optimise(self, solver_name="CBC", save_debug=False, save_results=False, results_dir=None):
        """
        Function called to run the optimisation solver. Files for decision variables and constraints are generated if
        save_debug_files is set as True. The generated optimisation results are saved in the planner module object.

        Parameters
        ----------
        save_debug_files : bool
                Save the decision variable and constraint files to local directory. The default is True.
        """
        svr = solver(solver_package=solver_name)

        # Save format
        if save_debug | save_results:
            # Check if results_dir is not None
            if (results_dir is None) or (not os.path.isdir(results_dir)):
                print("create results dir")
                if not os.path.isdir('results'):
                    os.mkdir('results')
                results_dir = os.path.join(os.getcwd(),'results')

            # Get timestamp for saving files
            timestamp = dt.strftime(dt.now(), "%Y_%m_%d_%H_%M")

            if save_debug:
                debug_fld = "NEMGLO_debug_" + timestamp
                sim_debug_dir = os.path.join(results_dir, debug_fld)
                os.mkdir(sim_debug_dir)

            if save_results:
                results_fld = "NEMGLO_results_" + timestamp
                sim_results_dir = os.path.join(results_dir, results_fld)
                os.mkdir(sim_results_dir)

    
        # Decision Variables
        decision_vars = pd.concat(self._var)
        decision_vars = decision_vars.droplevel(1, axis=0)
        decision_vars.index.name = 'variable_name'
        decision_vars.reset_index(inplace=True)
        svr.add_input_variables(input_variables=decision_vars)
        if save_debug:
            decision_vars.to_csv(os.path.join(sim_debug_dir, timestamp+"-DecisionVariables.csv"))

    # Objective Function Costs
        if self._objective_cost:
            obj_costs = pd.concat(self._objective_cost)
            obj_costs = obj_costs.droplevel(1, axis=0)
            obj_costs.index.name = 'variable_name'
            obj_costs.reset_index(inplace=True)
            svr.add_objective_function(objective_costs=obj_costs)
            if save_debug:
                obj_costs.to_csv(os.path.join(sim_debug_dir, timestamp+"-ObjectiveCosts.csv"))
        else:
            obj_costs = None

    # Constraint LHS
        if self._constr_lhs:
            constr_lhs = pd.concat(self._constr_lhs)
            constr_lhs = constr_lhs.droplevel(1, axis=0)
            constr_lhs.index.name = 'constraint_name'
            constr_lhs.reset_index(inplace=True)
            if save_debug:
                constr_lhs.to_csv(os.path.join(sim_debug_dir, timestamp+"-LHS.csv"))
        else:
            constr_lhs = None

    # Constraint RHS
        if self._constr_rhs:
            constr_rhs = pd.concat(self._constr_rhs)
            constr_rhs = constr_rhs.droplevel(1, axis=0)
            constr_rhs.index.name = 'constraint_name'
            constr_rhs.reset_index(inplace=True)
            svr.add_constraints(constr_lhs, constr_rhs)

            if save_debug:
                constr_rhs.to_csv(os.path.join(sim_debug_dir, timestamp+"-RHS.csv"))
        else:
            constr_rhs = None

    # Constraint RHS Dynamic
        if self._constr_rhs_dynamic:
            constr_rhsD = pd.concat(self._constr_rhs_dynamic)
            constr_rhsD = constr_rhsD.droplevel(1, axis=0)
            constr_rhsD.index.name = 'constraint_name'
            constr_rhsD.reset_index(inplace=True)
            svr.add_constraints_dynamic(constr_lhs, constr_rhsD)

            if save_debug:
                constr_rhsD.to_csv(os.path.join(sim_debug_dir, timestamp+"-RHS_dynamic.csv"))
        else:
            constr_rhsD = None

    # Big M constraint sets
        if self._constr_bigM:
            constr_bigM = pd.concat(self._constr_bigM)
            constr_bigM = constr_bigM.droplevel(1, axis=0)
            constr_bigM.index.name = 'constraint_name'
            constr_bigM.reset_index(inplace=True)
            svr.add_constraints_bigM(constr_lhs, constr_bigM)

            if save_debug:
                constr_bigM.to_csv(os.path.join(sim_debug_dir, timestamp+"-BigM.csv"))
        else:
            constr_bigM = None

    # SOS Type 2 set
        if self._sos_2:
            for sos_name in self._sos_2.keys():
                data = self._sos_2[sos_name]
                svr.add_sos_type_2(sos_name=sos_name,
                                    weights_df=data['weights'],
                                    x_samples=data['x_samples'],
                                    y_samples=data['y_samples'],
                                    xvar_df=data['x_vars'],
                                    yvar_df=data['y_vars'])

        # Concat RHS IDs
        all_rhs = pd.concat(
            [constr_rhs, constr_rhsD, constr_bigM], ignore_index=True)
        # Execute Optimiser
        # result_solver, removed_constrs = svr.optimise()
        result_solver, removed_constrs = svr.optimise()

        if removed_constrs:
            print("!Could not resolve constraints: {}".format(removed_constrs))
            rhs_problem_constrs = all_rhs['constraint_id'].isin([int(item) for item in removed_constrs])
            print(all_rhs[rhs_problem_constrs])
            all_rhs = all_rhs[~rhs_problem_constrs]

        self._out_vars = svr.get_decision_variable_vals(result_solver, decision_vars)
        self._out_constr = svr.get_constraint_slack(result_solver, all_rhs)
        """ need to remove searching for missing constraint"""

        # Optimisation outputs
        json_data, json_data['vars'], json_data['vars']['timeseries'], json_data['vars']['static'], json_data['inputtraces'] = {}, {}, {}, {}, {}
        #json_data['vars']['timeseries']['time'] = pd.Series(self._timeseries).to_json(orient='values')
        for variable in self._out_vars['variable_name'].unique():
            output = self._format_out_vars_timeseries(variable)

            if variable in ['h2_production','h2e_load']:
                json_data['vars']['timeseries'][variable] = output['value'].to_json(orient='values')
            else:
                json_data['vars']['static'][variable] = output['value'].to_json(orient='values')

            if save_results:
                output.to_csv(os.path.join(sim_results_dir,timestamp+"-"+variable+".csv"))

        # Input traces
        prc_output = pd.DataFrame({'time': self._timeseries,'interval':range(0,len(self._timeseries)),'value':self._prices})

        # if self._emissions:
        # 	em_output = pd.DataFrame({'time': self._timeseries,'interval':range(0,len(self._timeseries)),'value':self._emissions[0]._trace})

        json_data['inputtraces']['prices'] = prc_output.to_json()
        if save_results:
            prc_output.to_csv(os.path.join(sim_results_dir,timestamp+"-energy_price.csv"))
            if self._emissions:
                em_output.to_csv(os.path.join(sim_results_dir,timestamp+"-grid_emissions.csv"))

        # Write to file
        if save_results:
            with open(os.path.join(sim_results_dir,'outputs.json'), 'w') as json_file:
                json.dump(json_data, json_file)


        # Save costs to file
        if save_results:
            real_costs = self._sum_cost_components(cost_components='real')
            real_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_REAL.csv"))

            shadow_costs = self._sum_cost_components(cost_components='shadow')
            shadow_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_SHADOW.csv"))

            all_costs = self._sum_cost_components(cost_components='all')
            all_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_ALL.csv"))

            for variable in self._objective_cost:
                output = self._compute_cost_component(variable)
                output.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_"+variable+".csv"))

        # # Save all results to file
        # if save_results:
        #     # Optimiser Variables
        #     for variable in self._out_vars['variable_name'].unique():
        #         output = self._format_out_vars_timeseries(variable)
        #         output.to_csv(os.path.join(sim_results_dir,timestamp+"_"+variable+".csv"))

        #     # Input Variables
        #     prc_output = pd.DataFrame({'time': self._timeseries,'interval':range(0,len(self._timeseries)),'value':self._prices})
        #     prc_output.to_csv(os.path.join(sim_results_dir,timestamp+"_energy_price.csv"))

        return result_solver


    def get_load(self):
        """
        Retrieve the optimisation result for each interval load variable.

        Returns
        -------
        pd.DataFrame
                Returns dataframe containing timestamped and interval numbered load results.
        """
        elec = self._components['Electrolyser']
        prefix = elec[0]._id
        return self._format_out_vars_timeseries(prefix+'-mw_load')

    def get_production(self):
        """
        Retrieve the optimisation result for each interval h2 production variable.

        Returns
        -------
        pd.DataFrame
                Returns dataframe containing timestamped and interval numbered production results.
        """
        elec = self._components['Electrolyser']
        prefix = elec[0]._id
        return self._format_out_vars_timeseries(prefix+'-h2_produced')

    def get_storage_fill(self):
        """
        Retrieve the optimisation result for each interval h2 stored variable.

        Returns
        -------
        pd.DataFrame
                Returns dataframe containing timestamped and interval numbered stored results.
        """
        elec = self._components['Electrolyser']
        prefix = elec[0]._id
        return self._format_out_vars_timeseries(prefix+'-h2_stored')

    def get_impact_emissions(self):
        return self._format_out_vars_timeseries('impact_emissions')

    def get_vre_availability(self, identifier):
        """
        Retrieve the optimisation result for each interval vre availability for the specified DUID.

        Parameters
        ----------
        duid : str
                The DUID for which the vre availability should be returned for.

        Returns
        -------
        pd.DataFrame
                Returns dataframe containing timestamped and interval numbered vre availability results.
        """
        gens = self._components['Generator']
        specific_gen = [i for i in gens if i._id == identifier]
        prefix = specific_gen[0]._id
        return self._format_out_vars_timeseries(prefix+'-vre_avail')

    def get_vre_capacity_factor(self, identifier):
        """
        Retrieve the optimisation result for the average capacity factor over the simulation period for the specified
        DUID.

        Parameters
        ----------
        duid : str
                The DUID for which the capacity factor should be returned for.

        Returns
        -------
        float
                The capacity factor value corresponding to the specified VRE generator.
        """
        gens = self._components['Generator']
        specific_gen = [i for i in gens if i._id == identifier]
        return round(specific_gen[0]._trace[specific_gen[0]._duid].mean(), 3) * 100


    def get_total_consumption(self):
        """
        Retrieve the optimisation result for the total energy consumed by the load over the simulation period.

        Returns
        -------
        float
                The total energy consumption in MWh of the load.
        """
        elec = self._components['Electrolyser']
        prefix = elec[0]._id
        return float(self._out_vars[self._out_vars['variable_name'] == prefix+'-mw_load_sum'].value) * (self._intlength/60)


    def get_total_h2_production(self):
        """
        Retrieve the optimisation result for the total h2 produced by the load over the simulation period.

        Returns
        -------
        float
                The total energy consumption in MWh of the load.
        """
        elec = self._components['Electrolyser']
        prefix = elec[0]._id
        return float(self._out_vars[self._out_vars['variable_name'] == prefix+'-h2_produced_sum'].value)


    def get_load_capacity_factor(self):
        """
        Retrieve the optimisation result for the average capacity factor of the load over the simulation period.

        Returns
        -------
        load_capfac : float
                The capacity factor value of the electrolyser load.
        """
        elec = self._components['Electrolyser']
        load_mwh = self.get_total_consumption()
        load_capfac = load_mwh / (elec[0]._capacity * self._n * (self._intlength/60))
        return round(load_capfac,3) * 100


    def get_rec_on_interval(self, as_mwh=True):
        # Get REC series
        rec_df = self.get_load()[['interval','time','value']]
        rec_df.rename(columns={'value': str(self._components['Electrolyser'][0]._id) + ' RECs surrendered'}, \
            inplace=True)
        for series in [item for item in self._var.keys() if (item.__contains__('rec') and not item.__contains__('sum'))\
            or item.__contains__('vre_avail')]:
            res_set = self._format_out_vars_timeseries(series)
            res_set.loc[:,'value'] = res_set['value'].round(4)
            rec_df = rec_df.merge(res_set, on=['interval','time'], how='left')
            rec_df = rec_df.rename(columns={'value': series})

        if as_mwh:
            # Convert to MWh
            index_filt = rec_df.columns[~rec_df.columns.isin(['interval','time'])]
            rec_df.loc[:,index_filt] = np.ceil(rec_df[index_filt].mul((self._intlength/60)))

        # Rename REC columns
        for col in rec_df.columns:
            if col.__contains__(self._id) & col.__contains__('buy'):
                rec_df.rename(columns={col: 'Market RECs bought'}, inplace=True)
            elif col.__contains__(self._id) & col.__contains__('sell'):
                rec_df.rename(columns={col: 'Market RECs sold'}, inplace=True)
            elif 'Generator' in self._components:
                for gen in self._components['Generator']:
                    if col.__contains__(gen._id):
                        rec_df.rename(columns={col: '{} RECs acquired'.format(gen._id)}, inplace=True)

        return rec_df

    def get_rec_summary(self):
        rec_df = self._out_vars[(self._out_vars['variable_name'].str.contains('rec')) & \
                                (self._out_vars['interval'].isnull())]
        rec_df = rec_df.pivot_table(values='value',columns='variable_name')
        rec_df.columns.name = None
        rec_df = rec_df.round(4)
        rec_df = rec_df.mul((self._intlength/60)) # MWh
        rec_df = np.ceil(rec_df)

        # Add Electrolyser RECs surrendered
        elec_data = np.ceil(round(self.get_total_consumption(),4))
        rec_df.insert(0, str(self._components['Electrolyser'][0]._id) + ' RECs surrendered', \
            elec_data)

        # Rename REC columns
        for col in rec_df.columns:
            if col.__contains__(self._id) & col.__contains__('buy'):
                rec_df.rename(columns={col: 'Market RECs bought'}, inplace=True)
            elif col.__contains__(self._id) & col.__contains__('sell'):
                rec_df.rename(columns={col: 'Market RECs sold'}, inplace=True)
            elif 'Generator' in self._components:
                for gen in self._components['Generator']:
                    if col.__contains__(gen._id):
                        rec_df.rename(columns={col: '{} RECs acquired'.format(gen._id)}, inplace=True)
        return rec_df


    def get_lgc_summary(self):
        """
        Retrieve the optimisation result for LGCs received and surrendered by source.

        Returns
        -------
        lgc_summary : pd.DataFrame
                Returns a dataframe summarising the amount of LGCs received (and surrendered) in MWh terms, mapping to
                each source.
        """
        lgc_summary = pd.DataFrame(
            columns=['LGC_source', 'MWh_volume', 'percentage'])
        if self._out_vars['variable_name'].isin(['ppa_lgc_sum_1']).any():
            lgc_summary = pd.concat(
                [lgc_summary, self._format_out_lgc_stats('ppa_lgc_sum_1')])
        if self._out_vars['variable_name'].isin(['ppa_lgc_sum_2']).any():
            lgc_summary = pd.concat(
                [lgc_summary, self._format_out_lgc_stats('ppa_lgc_sum_2')])
        if self._out_vars['variable_name'].isin(['lgc_from_market']).any():
            lgc_summary = pd.concat(
                [lgc_summary, self._format_out_lgc_stats('lgc_from_market')])
        return lgc_summary


    def get_costs(self, exclude_shadow=True):
        output = pd.DataFrame()
        for variable in self._objective_cost:
            result = self._compute_cost_component(variable)
            result.insert(0,'variable_name',variable)
            output = pd.concat([output, result],ignore_index=True)

        table = output.pivot_table(index=['time','interval'], columns='variable_name', \
            values='value')
        table.insert(0,'total_cost',table.sum(axis=1))
        table = table.reset_index()
        table.columns.name = None

        if exclude_shadow:
            for col in table.columns:
                if any([col.__contains__(shadow_cost) for shadow_cost in cost_variables_shadow]):
                    table.drop(col, axis=1, inplace=True)
        return table

    # def get_costs_real(self):

    # 				real_costs = self._sum_cost_components(cost_components='real')
    # 		real_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_REAL.csv"))

    # 		shadow_costs = self._sum_cost_components(cost_components='shadow')
    # 		shadow_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_SHADOW.csv"))

    # 		all_costs = self._sum_cost_components(cost_components='all')
    # 		all_costs.to_csv(os.path.join(sim_results_dir,timestamp+"-cost_sum_ALL.csv"))

            

    def _compute_cost_component(self, var_name):
        cost_df = self._objective_cost[var_name]
        cost_df = cost_df.merge(self._out_vars[['variable_id','interval','value']], on=['variable_id','interval'])
        cost_df.rename(columns={'value':'variable_value'},inplace=True)
        cost_df.insert(3,'value', cost_df['cost'] * cost_df['variable_value'])

        if len(cost_df['interval']) == self._n:
            cost_df.insert(0,'time', self._timeseries)
        else:
            cost_df.insert(0,'time', np.nan)

        return cost_df[['time','interval','variable_value','cost','value']].reset_index(drop=True)


    def _sum_cost_components(self, cost_components):
        df = pd.DataFrame({'interval':range(0,len(self._timeseries))})

        if cost_components == 'real':
            filt_list = cost_variables_realized
        elif cost_components == 'shadow':
            filt_list = cost_variables_shadow
        elif cost_components == 'all':
            filt_list = cost_variables_realized + cost_variables_shadow
        else:
            raise Exception("Invalid cost_components argument. Must be one of `real`, `shadow`, `all`")

        for var_name in self._objective_cost:
            if any([var_name.__contains__(filt_list[i]) for i in range(len(filt_list))]):
                cost_df = self._compute_cost_component(var_name)
                df = df.merge(cost_df[['interval','value']], on=['interval'], how='left')
                df.rename(columns={'value': var_name}, inplace=True)

        df.set_index('interval', inplace=True)
        df['value'] = df.sum(axis=1)
        df.reset_index(inplace=True)
        df.insert(0,'time', self._timeseries)

        return df[['time','interval','value']]


    def _format_out_vars_timeseries(self, var_name):
        sim_times = pd.DataFrame({'interval': range(self._n), 'time': self._timeseries})

        result = self._out_vars[self._out_vars['variable_name'] == var_name][[
            'interval', 'value']]
        if (len(result) > 1) & (result['interval'].values[0] != None):
            result = result.sort_values('interval')
            result = result.merge(sim_times, on='interval', how='right')
        return result.reset_index(drop=True)

    def _format_out_lgc_stats(self, component_name):
        name_map = {'ppa_lgc_sum_1': self._vre_det['DUID'][1], 'ppa_lgc_sum_2': self._vre_det['DUID'][2],
                    'lgc_from_market': 'Market'}
        total_mwh = self.get_total_consumption()
        value = float(self._out_vars[self._out_vars['variable_name']
                      == component_name].value) * (self._intlength/60)
        percent = (value / total_mwh) * 100
        lgc_component = pd.DataFrame({'LGC_source': [name_map[component_name]], 'MWh_volume': [int(np.floor(value))],
                                      'percentage': [percent]})
        return lgc_component


    def _find_component_class(self, component_class):
        return [c for c in self._components if c.__class__.__name__ == component_class]